#!/usr/bin/env python
"""
The abirun.py script allows the user to submit the calculations contained in an AbiPy `Flow`.
It provides a command line interface as well graphical interfaces.
"""
from __future__ import annotations

import sys
import os
import argparse
import shlex
import time
import platform
import tempfile
import abipy.flowtk as flowtk
import abipy.abilab as abilab
import abipy.tools.cli_parsers as cli

from pprint import pprint
from collections import defaultdict, OrderedDict
from socket import gethostname
from monty import termcolor
from monty.functools import prof_main
from monty.termcolor import cprint, colored, get_terminal_size
from monty.string import boxed, make_banner
from abipy.tools import duck
from abipy.flowtk import Status
from abipy.flowtk.flows import Flow


def straceback() -> str:
    """Returns a string with the traceback."""
    import traceback
    return traceback.format_exc()


def parse_strings(s: str):
    """Parse comma separated values. Return None if s is None."""
    return s.split(",") if s is not None else s


def flowdir_wname_tname(dirname: str):
    """"
    Given a initial directory `dirname` containing a node of the `Flow`,
    this function locates the directory of the flow (e.g. the directory with the pickle file)
    and returns the name of the work and/or of the node.

    Return: flowdir, wname, tname

    where flowdir is the directory containing the pickle file,
    wname and tname are the basenames of the work/task.

    If dirname contains the pickle file we have (wname, tname) == (None, None)
    If dirname is a work --> wname is it's basename and tname is None
    If dirname is a task --> os.path.join(flowdir, wname, tname) == task.workdir.
    """
    if dirname is None: dirname = os.getcwd()
    dirname = os.path.abspath(dirname)
    if os.path.exists(os.path.join(dirname, flowtk.Flow.PICKLE_FNAME)):
        return dirname, None, None

    # Handle works or tasks.
    head = dirname
    wname, tname = None, None
    for i in range(2):
        head, tail = os.path.split(head)
        if i == 0: tail_1 = tail
        if os.path.exists(os.path.join(head, flowtk.Flow.PICKLE_FNAME)):
            if i == 0:
                # We have a work: /root/flow_dir/w[num]
                wname = tail
            if i == 1:
                # We have a task: /root/flow_dir/w[num]/t[num]
                wname = tail
                tname = tail_1

            #print("wname", wname, "tname", tname)
            return head, wname, tname

    raise RuntimeError("Cannot locate flowdir from %s" % dirname)


def select_nids(flow: Flow, options) -> set:
    """
    Return the set of node ids selected by the user via the command line interface.
    """
    #print("options.nids:", options.nids)
    task_ids = [task.node_id for task in
                flow.select_tasks(nids=options.nids, wslice=options.wslice, task_class=options.task_class)]

    # Have to add the ids of the works containing the tasks.
    if options.nids is not None:
        work_ids = [work.node_id for work in flow if work.node_id in options.nids]
    else:
        work_ids = [work.node_id for work in flow]

    return set(work_ids + task_ids)


def cli_abiopen(options, filepath: str):
    """Code taken from abiopen.py."""
    # TODO: One should have a single API but make_and_open_notebook are different so
    # for the time being we use two different versions.
    filepath = options.flowdir
    if filepath is None:
        raise RuntimeError("filepath must be specified")

    if not os.path.exists(filepath):
        raise RuntimeError("%s: no such file" % filepath)

    if not options.notebook:
        # Start ipython shell with namespace
        abifile = abilab.abiopen(filepath)
        if options.print:
            if hasattr(abifile, "to_string"):
                print(abifile.to_string(verbose=options.verbose))
            else:
                print(abifile)
            return 0

        import IPython
        # Use embed because I don't know how to show a header with start_ipython.
        IPython.embed(header="The Abinit file is associated to the `abifile` variable.\nTry `print(abifile)`")

    else:
        # Call specialized method if the object is a NotebookWriter
        # else generate simple notebook by calling `make_and_open_notebook`
        cls = abilab.abifile_subclass_from_filename(filepath)
        if hasattr(cls, "make_and_open_notebook"):
            if hasattr(cls, "__exit__"):
                with abilab.abiopen(filepath) as abifile:
                    return abifile.make_and_open_notebook(foreground=options.foreground)
            else:
                abifile = abilab.abiopen(filepath)
                return abifile.make_and_open_notebook(foreground=options.foreground)
        else:
            raise TypeError("Object ot type `%s` does not provide make_and_open_notebook method" % str(cls))
            #return make_and_open_notebook(options)


def flow_debug_reset_tasks(flow: Flow, nids=None, verbose=0) -> None:
    """
    Analyze error files produced by reset tasks for possible error messages

    Args:
        nids: List of node identifiers. By defaults all nodes that have been resetted are analyzed.
        verbose: Verbosity level.
    """
    # TODO: Improve implementation.
    ntasks = 0
    nrows, ncols = get_terminal_size()
    for task in flow.select_tasks(nids=nids):
        # See task.reset_from_scratch
        if task.num_restarts == 0: continue
        reset_dir = os.path.join(task.workdir, "_reset")
        reset_file = os.path.join(reset_dir, "_counter")
        if not os.path.isdir(reset_dir) and not os.path.isfile(reset_file):
            continue

        ntasks += 1
        with open(reset_file, "rt") as fh:
            num_reset = int(fh.read())

        for i in range(num_reset):
            #("output_file", "log_file", "stderr_file", "qout_file", "qerr_file", "mpiabort_file")
            for fname in ("stderr_file", "qerr_file", "mpiabort_file"):
                path = os.path.join(reset_dir, fname + "_" + str(i))
                with open(path, "rt") as fh:
                    s = fh.read()
                    if not s: continue
                    print(2 * "\n")
                    print(make_banner(os.path.relpath(path), width=ncols, mark="="))
                    cprint(s, color="red")
                    print(2 * "\n")

    print("Number of tasks analyzed: %d" % ntasks)


def flow_watch_status(flow: Flow, delay=5, nids=None, verbose=0, func_name="show_func"):
    """
    Enter an infinite loop and delay execution for the given number of seconds. (default: 5 secs).

    Args:
        delay: delay execution for the given number of seconds. (default: 5 secs).
        nids: List of node identifiers. By defaults all nodes that have been resetted are analyzed.
        verbose: Verbosity level.
        func_name: Name of the function used to show the status of the flow.
    """
    cprint("Entering infinite loop (delay: %d s). Only changes are shown\nPress <CTRL+C> to exit" %
           delay, color="magenta", end="", flush=True)

    show_func = getattr(flow, func_name)
    assert callable(show_func)

    # Total counter and dicts used to detect changes.
    tot_count = 0
    before_task2stat, now_task2stat = {}, {}
    # Progressbar setup
    from tqdm import tqdm
    pbar, pbar_count, pbar_total = None, 0, 100

    exit_code = 0
    def exit_now():
        """
        Function used to test if we have to exit from the infinite loop below.
        Return: != 0 if we must exit. > 0 if some error occurred.
        """
        if flow.all_ok:
            cprint("Flow reached all_ok", "green")
            return -1
        if any(st.is_critical for st in before_task2stat.values()):
            cprint(boxed("Found tasks with critical status"), "red")
            return 1
        return 0

    try:
        while True:
            tot_count += 1
            flow.check_status()

            # Here I test whether there's been some change in the flow
            # before printing the status table.
            # Note that the flow in memory could not correspond to the one that
            # is being executed by the scheduler. This is the reason why we
            # reload it when we reach pbar_count.
            if tot_count == 1:
                for task in flow.iflat_tasks(nids=nids):
                    before_task2stat[task] = task.status
            else:
                for task in flow.iflat_tasks(nids=nids):
                    now_task2stat[task] = task.status

                if (len(before_task2stat) == len(now_task2stat) and
                    all(now_task2stat[t] == before_task2stat[t] for t in now_task2stat)):
                    # In principle this is not needed but ...
                    exit_code = exit_now()
                    if exit_code: break

                    # Progress bar section.
                    if pbar is None:
                        print("No change detected in the flow. Won't print status table till next change...")
                        pbar = tqdm(total=pbar_total)

                    if pbar_count <= pbar_total:
                        pbar_count += 1
                        pbar.update(1)
                    else:
                        pbar_count = 0
                        pbar.close()
                        pbar = tqdm(total=pbar_total)
                        flow.reload()

                    time.sleep(delay)
                    continue

                # copy now --> before
                before_task2stat = now_task2stat.copy()

            # Print status table. Exit if success or critical errors.
            print(2*"\n" + time.asctime() + "\n")
            show_func(verbose=verbose, nids=nids)
            # Add summary table to status table.
            if show_func is flow.show_status: flow.show_summary()

            exit_code = exit_now()
            if exit_code: break
            time.sleep(delay)

        # Print status table if something bad happened.
        if exit_code == 1:
            flow.show_status()

    except KeyboardInterrupt:
        cprint("Received KeyboardInterrupt from user\n", "yellow")


def get_epilog() -> str:
    usage = """\

Usage example:

###########
# Execution
###########

  abirun.py FLOWDIR rapid                 => Keep repeating, stop when no task can be executed.
  abirun.py FLOWDIR scheduler             => Execute flow with the scheduler.
  abirun.py FLOWDIR status                => Show status table.
  abirun.py FLOWDIR events                => Print ABINIT events (Warnings/Errors/Comments) found in log files.
  abirun.py FLOWDIR history               => Print Task histories.
  abirun.py FLOWDIR cancel                => Cancel jobs in the queue.
  abirun.py FLOWDIR debug                 => Analyze error files and log files for possible error messages.
  abirun.py FLOWDIR corrections           => Show AbiPy corrections performed at runtime.
  abirun.py FLOWDIR handlers              => Show event handlers installed in the flow.

##########
# Analysis
##########

  abirun.py FLOWDIR inputs                => Print input files.
  abirun.py FLOWDIR abivars -vn ecut,nband  => Print table with these input variables.
  abirun.py FLOWDIR structures            => Compare input/output structures of the tasks.
  abirun.py FLOWDIR ebands -t NscfTask    => Print table with electronic properties computed in NscfTask
  abirun.py FLOWDIR hist -p               => Print table with last iteration in hist files and plot results.
  abirun.py FLOWDIR cycles -p             => Print (and plot) SCF/relaxation cycles extracted from the output of the tasks.
  abirun.py FLOWDIR robot hist            => Build robot for HIST files, start ipython shell to interact with the robot.
  abirun.py FLOWDIR dims                  => Print table with dimensions extracted from the output of the tasks.
  abirun.py FLOWDIR group -g task_class   => Print table with node ids and tasks grouped by class
  abirun.py FLOWDIR inspect               => Call matplotlib to inspect the tasks.
  abirun.py FLOWDIR tail                  => Use Unix tail to follow the main output files of the flow.
  abirun.py FLOWDIR deps                  => Show task dependencies.
  abirun.py FLOWDIR listext GSR SIGRES    => Show output files with the given extension(s).
  abirun.py FILE abiopen                  => Open FILE with abiopen. Same interface as abiopen.py

###############
# Miscelleanous
###############

  abirun.py FLOWDIR ipython               => Open flow in ipython terminal.
  abirun.py FLOWDIR notebook              => Generate jupyter notebook.
  abirun.py FLOWDIR networkx              => Plot flow graph with networkx.
  abirun.py FLOWDIR graphviz              => Plot flow graph with graphviz
                                            (can also select tasks/works and show files with -d).
  abirun.py abibuild                      => Show ABINIT build information and exit.

###############
# Documentation
###############

  abirun.py FLOWDIR doc_manager slurm     => Document the TaskManager options availabe for Slurm.
  abirun.py . doc_manager script          => Show the job script that will be produced with the current settings.
  abirun.py . doc_scheduler               => Document the options available in scheduler.yml.
"""

    notes = """\

Notes:

    If FLOWDIR is not given, abirun.py automatically selects the database located within
    the working directory. An Exception is raised if multiple databases are found.

    Note, moreover, that one can replace FLOWDIR with the directory of a work/task
    to make the command operate on this node of the flow without having to specify the node ids with --nids.
    For example, to have the list of events of the task in `FLOWDIR/w0/t1` use:

        abirun.py FLOWDIR/w0/t1 events

    instead of:

        abirun.py FLOWDIR events -n 123

    where 123 is the node identifier associated to w0/t1.

    To start the scheduler with a time interval of 30 seconds, use:

        nohup abirun.py FLOWDIR scheduler -s 30 > sched.log 2> sched.err

    Alternatively one can specify the scheduler options via the `scheduler.yml` file.
    Remember that AbiPy will first look for `scheduler.yml` and `manager.yml` files
    in the current working directory and then inside $HOME/.abinit/abipy/

    Use `abirun.py --help` for help and `abirun.py COMMAND --help` to get the documentation for `COMMAND`.
    Use `-v` to increase verbosity level (can be supplied multiple times e.g -vv).
"""

    developers = """\

############
# Developers
############

    abirun.py prof ABIRUN_ARGS               => to profile abirun.py
    abirun.py tracemalloc ABIRUN_ARGS        => to trace memory blocks allocated by Python"""

    return notes + usage # + developers


def get_parser(with_epilog=False):

    def parse_nids(s):
        """parse nids argument"""
        if s is None: return s
        try:
            if "," in s:
                return [int(t) for t in s.split(",")]
            else:
                # Convert string to slice and return list.
                s = duck.as_slice(s)
                if s.stop is None: raise argparse.ArgumentTypeError("stop must be specified")
                return list(range(s.start, s.stop, s.step))
        except Exception:
            raise argparse.ArgumentTypeError(
                    "Invalid nids string %s\n Expecting None or int or comma-separated integers or slice sintax" % s)

    def parse_wslice(s):
        s = duck.as_slice(s)
        if s is None: return s
        if s.stop is None: raise argparse.ArgumentTypeError("stop must be specified")
        return s

    # Parent parser for commands that need to know on which subset of tasks/workflows we have to operate.
    # wslide and nids are mutually exclusive.
    flow_selector_parser = argparse.ArgumentParser(add_help=False)
    group = flow_selector_parser.add_mutually_exclusive_group()
    group.add_argument("-n", '--nids', default=None, type=parse_nids, help=(
        "Node identifier(s) used to select the task. Accept single integer, comma-separated list of integers or python slice.\n"
        "Use `status` command to get the node ids.\n"
        "Examples: --nids=12 --nids=12,13,16 --nids=10:12 to select 10 and 11 (slice syntax), --nids=2:5:2 to select 2,4."
        ))

    group.add_argument("-w", '--wslice', default=None, type=parse_wslice,
        help=("Select the list of works to analyze (python syntax for slices): "
              "Examples: --wslice=1 to select the second workflow, --wslice=:3 for 0,1,2, "
              "--wslice=-1 for the last workflow, --wslice::2 for even indices."))
    group.add_argument("-S", '--task-status', default=None, type=Status.as_status,
        help="Select only the tasks with the given status. Default: None i.e. ignored. Possible values: %s." %
             Status.all_status_strings())
    group.add_argument("-t", '--task-class', type=str, default=None,
        help="Select only tasks with the given class e.g. `-t NscfTask`.")

    # Parent parser for common options.
    copts_parser = argparse.ArgumentParser(add_help=False)
    copts_parser.add_argument('-v', '--verbose', default=0, action='count', # -vv --> verbose=2
        help='verbose, can be supplied multiple times to increase verbosity.')

    copts_parser.add_argument('--no-colors', default=False, action="store_true", help='Disable ASCII colors.')
    copts_parser.add_argument('--no-logo', default=False, action="store_true", help='Disable AbiPy logo.')
    copts_parser.add_argument('--loglevel', default="ERROR", type=str,
        help="Set the loglevel. Possible values: CRITICAL, ERROR (default), WARNING, INFO, DEBUG.")
    copts_parser.add_argument('--remove-lock', default=False, action="store_true",
        help="Remove the lock on the pickle file used to save the status of the flow.")

    # Build the main parser.
    parser = argparse.ArgumentParser(epilog=get_epilog() if with_epilog else "",
                                     formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument('flowdir', nargs="?", help=("File or directory containing the ABINIT flow/work/task. "
                                                    "If not given, the flow in the current workdir is selected."))
    parser.add_argument('-V', '--version', action='version', version=abilab.__version__)

    # Create the parsers for the sub-commands
    subparsers = parser.add_subparsers(dest='command', help='sub-command help', description="Valid subcommands")

    # Subparser for single command.
    p_single = subparsers.add_parser('single', parents=[copts_parser], help="Run single task and exit.")

    # Subparser for rapid command.
    p_rapid = subparsers.add_parser('rapid', parents=[copts_parser], help="Run all tasks in rapidfire mode.")
    p_rapid.add_argument('-m', '--max-nlaunch', default=10, type=int,
        help="Maximum number of launches. default: 10. Use -1 for no limit.")

    # Subparser for scheduler command.
    p_scheduler = subparsers.add_parser('scheduler', parents=[copts_parser],
        help="Run all tasks with a Python scheduler. Requires scheduler.yml either in $PWD or ~/.abinit/abipy.")
    p_scheduler.add_argument('-w', '--weeks', default=0, type=int, help="Number of weeks to wait.")
    p_scheduler.add_argument('-d', '--days', default=0, type=int, help="Number of days to wait.")
    p_scheduler.add_argument('-hs', '--hours', default=0, type=int, help="Number of hours to wait.")
    p_scheduler.add_argument('-m', '--minutes', default=0, type=int, help="Number of minutes to wait.")
    p_scheduler.add_argument('-s', '--seconds', default=0, type=int, help="Number of seconds to wait.")

    # Subparser for status command.
    p_status = subparsers.add_parser('status', parents=[copts_parser, flow_selector_parser], help="Show status table.")
    p_status.add_argument('-d', '--delay', nargs="?", const=5, default=0, type=int,
        help="Enter an infinite loop and delay execution for the given number of seconds. (default: 5 secs).")
    p_status.add_argument('-s', '--summary', default=False, action="store_true",
        help="Print short version with status counters.")

    # Subparser for set_status command.
    p_set_status = subparsers.add_parser('set_status', parents=[copts_parser, flow_selector_parser],
        help="Change the status of the task. WARNING: Option for developers!")
    p_set_status.add_argument('new_status', help="New value of status. Possible values: %s." % Status.all_status_strings())

    # Subparser for cancel command.
    p_cancel = subparsers.add_parser('cancel', parents=[copts_parser, flow_selector_parser],
        help="Cancel the tasks in the queue. Not available if qtype is shell.")
    p_cancel.add_argument("-r", "--rmtree", action="store_true", default=False, help="Remove flow directory.")

    # Subparser for restart command.
    p_restart = subparsers.add_parser('restart', parents=[copts_parser, flow_selector_parser],
        help=("Restart the tasks of the flow. By default, only the task whose status==Unconverged are restarted. "
              "Use -S `status` and/or -n node_ids to select particular tasks."))

    # Subparser for reset command.
    p_reset = subparsers.add_parser('reset', parents=[copts_parser, flow_selector_parser],
        help="Reset the tasks of the flow with the specified status.")

    # Subparser for move command.
    p_move = subparsers.add_parser('move', parents=[copts_parser],
        help="Move the flow to a new directory and change the absolute paths.")
    p_move.add_argument('dest', nargs=1)

    # Subparser for open command.
    p_open = subparsers.add_parser('open', parents=[copts_parser, flow_selector_parser],
        help="Open files in $EDITOR, type `abirun.py FLOWDIR open --help` for help).")
    p_open.add_argument('what', nargs="?", default="o",
        choices=["i", "o", "f", "j", "l", "e", "q", "all"],
        help=r"""\
Specify the files to open. Possible choices:
    i ==> input_file
    o ==> output_file
    f ==> files_file
    j ==> job_file
    l ==> log_file
    e ==> stderr_file
    q ==> qout_file
    all ==> all files.
Default: o
""")

    # Subparser for abiopen.
    p_abiopen = subparsers.add_parser('abiopen', parents=[copts_parser], #, flow_selector_parser],
        help="Open FILE with abiopen. Same interface as abiopen.py")
    p_abiopen.add_argument('-nb', '--notebook', action='store_true', default=False, help="Open file in jupyter notebook")
    p_abiopen.add_argument('--foreground', action='store_true', default=False,
                        help="Run jupyter notebook in the foreground.")
    p_abiopen.add_argument('-p', '--print', action='store_true', default=False, help="Print python object and return.")

    # Subparser for abibuild
    p_abibuild = subparsers.add_parser('abibuild', parents=[copts_parser, flow_selector_parser],
        help="Show ABINIT build information and exit.")

    # Subparser for doc_scheduler
    p_docsched = subparsers.add_parser('doc_scheduler', parents=[copts_parser],
        help="Document the options available in scheduler.yml.")

    # Subparser for explain_(works|tasks)
    p_explain_works = subparsers.add_parser('explain_works', parents=[copts_parser, flow_selector_parser],
        help="Explain operations performed by Works")
    p_explain_tasks = subparsers.add_parser('explain_tasks', parents=[copts_parser, flow_selector_parser],
        help="Explain operations performed by Tasks")

    # Subparser for autoparal
    p_autoparal = subparsers.add_parser('autoparal', parents=[copts_parser, flow_selector_parser],
        help="Show autoparal configurations and optimal ones.")

    # Subparser for panel
    p_panel = subparsers.add_parser('panel', parents=[copts_parser, flow_selector_parser],
                                    help="Interact with the flow in the browser (requires panel package).")
    p_panel.add_argument("-pnt", "--panel-template", default="FastList", type=str,
                        help="Specify template for panel dasboard." +
                             "Possible values are: FastList, FastGrid, Golden, Bootstrap, Material, React, Vanilla." +
                             "Default: FastList"
                        )
    p_panel.add_argument('--no-browser', action='store_true', default=False,
                        help=("Start the bokeh server to serve the panel app "
                              "but don't open the app in the browser.\n"
                              "Use this option to connect remotely from localhost to the machine running the server"))
    p_panel.add_argument("--port", default=0, type=int, help="Allows specifying a specific port when serving panel app.")

    # Subparser for new_manager.
    p_new_manager = subparsers.add_parser('new_manager', parents=[copts_parser, flow_selector_parser],
        help="Change the TaskManager.")
    p_new_manager.add_argument("manager_file", default="", type=str, help="YAML file with the new manager.")

    # Subparser for tail.
    p_tail = subparsers.add_parser('tail', parents=[copts_parser, flow_selector_parser],
        help="Use unix tail to follow the main output files of the flow.")
    p_tail.add_argument('what_tail', nargs="?", type=str, default="o", choices=["o", "l", "e"],
        help="What to follow: `o` for output (default), `l` for logfile, `e` for stderr.")

    # Subparser for tail.
    p_timeit = subparsers.add_parser('timeit', parents=[copts_parser, flow_selector_parser],
        help="Extract timing data from Abinit output files.")

    # Subparser for qstat.
    # TODO: finalize the implementation
    #p_qstat = subparsers.add_parser('qstat', parents=[copts_parser], help="Show additional info on the jobs in the queue.")

    # Subparser for deps.
    p_deps = subparsers.add_parser('deps', parents=[copts_parser], help="Show dependencies.")

    # Subparser for robot.
    p_robot = subparsers.add_parser('robot', parents=[copts_parser, flow_selector_parser],
        help="Use a robot to analyze the results of multiple tasks (requires ipython).")
    p_robot.add_argument('robot_ext', nargs="?", type=str, default="GSR",
        help=("The file extension of the output file (case insensitive). "
              "Must be in: %s" % str(abilab.Robot.get_supported_extensions())))
    p_robot.add_argument('-nb', '--notebook', action='store_true', default=False, help="Generate jupyter notebook.")
    p_robot.add_argument('--foreground', action='store_true', default=False,
        help="Run jupyter notebook in the foreground.")
    p_robot.add_argument('-p', '--print-dataframe', action='store_true', default=False,
        help="Invoke robot.get_dataframe() and printa table to terminal.")

    # Subparser for cycles.
    p_cycles = subparsers.add_parser('cycles', parents=[copts_parser, flow_selector_parser],
        help=("Print self-consistent/relaxation cycles extracted from the output of the tasks."
              "Use `-t ScfTask` or `FLOWDIR/w0/` or `FLOWDIR/w0/t1/` to select tasks by class, by work or by workdir."))
    p_cycles.add_argument("-wok", "--exclude-ok-tasks", action='store_true', default=False,
        help="Exclude Tasks that have reached S_OK.")
    p_cycles.add_argument("-p", "--plot-mode", nargs="?", default=None, const="combiplot",
        choices=["combiplot", "slideshow"], #"gridplot",
        help=("Plot multiple cycles on the same figure if arg is specified. Use `-p` for gridplot. "
              "Use `-p slideshow` to iterate."))

    # Subparser for dims.
    p_dims = subparsers.add_parser('dims', parents=[copts_parser, flow_selector_parser],
        help="Print table with dimensions extracted from the output of the tasks.")

    # Subparser for inspect.
    p_inspect = subparsers.add_parser('inspect', parents=[copts_parser, flow_selector_parser],
        help="Call matplotlib to inspect the tasks (execute task.inspect method)")

    # Subparser for inputs.
    p_inputs = subparsers.add_parser('inputs', parents=[copts_parser, flow_selector_parser],
        help="Show the input files of the tasks.")
    p_inputs.add_argument("-vn", "--varnames", nargs="?", default=None, type=parse_strings,
        help="Comma-separated variable names. Can be used to print only these variables.")

    # Subparser for abivars.
    p_abivars = subparsers.add_parser('abivars', parents=[copts_parser, flow_selector_parser],
        help="Show pandas dataframe with Abinit input variables.")
    p_abivars.add_argument("-vn", "--varnames", required=True, type=parse_strings,
        help="Comma-separated variable names e.g. `-vn ecut,nband,ngkpt`.")

    # Subparser for structures command.
    p_structures = subparsers.add_parser('structures', parents=[copts_parser, flow_selector_parser],
        help="Compare input/output structures of the tasks. Print max force and pressure if available.")
    p_structures.add_argument("--what", type=str, default="io",
        help="'i' for input structures, 'o' for output, 'io' for both.")

    # Subparser for ebands command.
    p_ebands = subparsers.add_parser('ebands', parents=[copts_parser, flow_selector_parser],
        help="Compare electronic bands produced by the tasks.")
    p_ebands.add_argument("-p", "--plot-mode", nargs="?", default=None, const="gridplot",
        choices=["gridplot", "combiplot", "boxplot", "combiboxplot", "animate"],
        help="Plot multiple bands if arg is specified. Use -p for gridplot. Supports multiple formats.")

    # Subparser for hist command.
    p_hist = subparsers.add_parser('hist', parents=[copts_parser, flow_selector_parser],
        help="Compare HIST.nc files produced by the tasks.")
    p_hist.add_argument("-p", "--plot-mode", nargs="?", default=None, const="combiplot",
        choices=["gridplot", "combiplot"],
        help="Plot multiple HIST files if arg is present. Use -p for combiplot. Supports multiple formats.")

    # Subparser for manager.
    p_manager = subparsers.add_parser('doc_manager', parents=[copts_parser], help="Document the TaskManager options.")
    p_manager.add_argument("qtype", nargs="?", default=None, help=("Write job script to terminal if qtype='script' else "
        "document the qparams for the given QueueAdapter qtype e.g. slurm."))

    # Subparser for events.
    p_events = subparsers.add_parser('events', parents=[copts_parser, flow_selector_parser],
        help="Show ABINIT events (error messages, warnings, comments).")
    #p_events.add_argument("-t", "event-type", default=)

    # Subparser for corrections.
    p_corrections = subparsers.add_parser('corrections', parents=[copts_parser, flow_selector_parser],
        help="Show AbiPy corrections performed at runtime.")

    # Subparser for history.
    p_history = subparsers.add_parser('history', parents=[copts_parser, flow_selector_parser], help="Show Node history.")
    p_history.add_argument("-m", "--metadata", action="store_true", default=False, help="Print history metadata.")
    p_history.add_argument("-f", "--full-history", action="store_true", default=False,
        help="Print full history set, including nodes with an empty history.")

    # Subparser for handlers.
    p_handlers = subparsers.add_parser('handlers', parents=[copts_parser],
        help="Show event handlers installed in the flow.")
    p_handlers.add_argument("-d", "--doc", action="store_true", default=False,
        help="Show documentation about all the handlers that can be installed.")

    # Subparser for notebook.
    p_notebook = subparsers.add_parser('notebook', parents=[copts_parser],
        help="Create and open an ipython notebook to interact with the flow.")
    p_notebook.add_argument('--foreground', action='store_true', default=False,
        help="Run jupyter notebook in the foreground.")

    # TODO:
    #parser.add_argument('--classic-notebook', "-cnb", action='store_true', default=False,
    #                    help="Use classic jupyter notebook instead of jupyterlab.")
    #parser.add_argument('--no-browser', action='store_true', default=False,
    #                    help=("Start the jupyter server to serve the notebook "
    #                          "but don't open the notebook in the browser.\n"
    #                          "Use this option to connect remotely from localhost to the machine running the kernel"))
    #parser.add_argument('--foreground', action='store_true', default=False,
    #                    help="Run jupyter notebook in the foreground.")

    # Subparser for ipython.
    p_ipython = subparsers.add_parser('ipython', parents=[copts_parser],
        help="Embed IPython. Useful for advanced operations or debugging purposes.")
    p_ipython.add_argument('--argv', nargs="?", default="", type=shlex.split,
        help="Command-line options passed to ipython. Must be enclosed by quotes. "
             "Example: --argv='--matplotlib=wx'")

    # Subparser for tar.
    p_tar = subparsers.add_parser('tar', parents=[copts_parser], help="Create tarball file.")
    p_tar.add_argument("-s", "--max-filesize", default=None,
        help="Exclude file whose size > max-filesize bytes. Accept integer or string e.g `1MB`.")
    p_tar.add_argument("-e", "--exclude-exts", default=None, type=parse_strings,
        help="Exclude file extensions. Accept string or comma-separated strings. Ex: -eWFK or --exclude-exts=WFK,GSR")
    p_tar.add_argument("-d", "--exclude-dirs", default=None, type=parse_strings,
        help="Exclude directories. Accept string or comma-separated strings. Ex: --exlude-dirs=indir,outdir")
    p_tar.add_argument("-l", "--light", default=False, action="store_true",
        help="Create light-weight version of the tarball for debugging purposes. Other options are ignored.")

    # Subparser for tricky.
    p_tricky = subparsers.add_parser('tricky', parents=[copts_parser],
        help=("Show tricky tasks i.e. tasks that have been restarted, "
              "launched more than once or tasks that have been corrected."))

    # Subparser for debug.
    p_debug = subparsers.add_parser('debug', parents=[copts_parser, flow_selector_parser],
        help="Analyze error files and log files for possible error messages.")

    # Subparser for debug_reset.
    p_debug_reset = subparsers.add_parser('debug_reset', parents=[copts_parser, flow_selector_parser],
        help="Analyze error files and log files produced by reset tasks for possible error messages.")

    # Subparser for reset_jobids.
    p_reset_jobids = subparsers.add_parser('reset_jobids', parents=[copts_parser, flow_selector_parser],
        help="Analyze error files and log files produced by reset tasks for possible error messages.")

    # Subparser for clone_task.
    #p_clone_task = subparsers.add_parser('clone_task', parents=[copts_parser, flow_selector_parser],
    #    help="Clone task, change input variables and add new tasks to the flow. Requires clone_task.py.")

    # Subparser for group.
    p_group = subparsers.add_parser('group', parents=[copts_parser, flow_selector_parser],
        help="Group tasks according to property.")
    p_group.add_argument("-g", '--groupby', default="status", type=str, choices=["status", "task_class"],
        help="Select the attribute used to group tasks. Default: status.")

    # Subparser for diff.
    p_diff = subparsers.add_parser('diff', parents=[copts_parser, flow_selector_parser],
        help="Compare files produced by two or three nodes.")
    p_diff.add_argument('what_diff', nargs="?", type=str, default="i",
        help="What to diff: `i` for input (default), `o` for output, `l` for logfile, `e` for stderr.")

    # Subparser for networkx.
    p_networkx = subparsers.add_parser('networkx', parents=[copts_parser],
        help="Draw flow and node dependencies with networkx package.")
    p_networkx.add_argument('--nxmode', default="status",
        help="Type of network plot. Possible values: `status`, `network`. Default: `status`.")
    p_networkx.add_argument('--edge-labels', action="store_true", default=False, help="Show edge labels.")

    # Subparser for graphviz.
    p_graphviz = subparsers.add_parser('graphviz', parents=[copts_parser],
        help=("Draw flow and node dependencies with graphviz package. Accept (FLOWDIR|WORKDIR|TASKDIR). "
             "See https://graphviz.readthedocs.io/."))
    p_graphviz.add_argument("-e", "--engine", type=str, default="automatic",
        help=("graphviz engine: ['dot', 'neato', 'twopi', 'circo', 'fdp', 'sfdp', 'patchwork', 'osage']. "
            "Default: automatic i.e. the engine is automatically selected. See http://www.graphviz.org/pdf/dot.1.pdf "
            "Use `conda install python-graphviz` or `pip install graphviz` to install the python package."))
    p_graphviz.add_argument("-d", '--dirtree', default=False, action="store_true",
        help='Visualize files and directories in workdir instead of tasks/works.')

    # Subparser for listext.
    p_listext = subparsers.add_parser('listext', parents=[copts_parser],
        help="List all the output files with the given extension that have been produced by the nodes.")
    p_listext.add_argument('listexts', nargs="*", default=[], help="List of Abinit file extensions. e.g DDB, GSR, WFK etc")

    return parser


def serve_kwargs_from_options(options) -> dict:

    #address = "localhost"
    if options.no_browser:
        print("""
Use:

    ssh -N -f -L localhost:{port}:localhost:{port} username@your_remote_cluster

for port forwarding.
""")

    import abipy.panels as mod
    assets_path = os.path.join(os.path.dirname(mod.__file__), "assets")

    return dict(
        debug=options.verbose > 0,
        show=not options.no_browser,
        port=options.port,
        static_dirs={"/assets": assets_path},
        #address=address,
        #websocket_origin="{address}:{port}",
    )



@prof_main
def main():

    def show_examples_and_exit(err_msg=None, error_code=1):
        """Display the usage of the script."""
        sys.stderr.write(get_epilog())
        if err_msg: sys.stderr.write("Fatal Error\n" + err_msg + "\n")
        sys.exit(error_code)

    parser = get_parser(with_epilog=True)

    # Parse command line.
    try:
        options = parser.parse_args()
    except Exception as exc:
        show_examples_and_exit(error_code=1)

    if not options.command:
        show_examples_and_exit(error_code=1)

    cli.set_loglevel(options.loglevel)

    if options.verbose > 2:
        print(options)

    # Documentation options that do not need a flow.
    # Print docs and exit immediately.
    if options.command == "doc_manager":
        # Document TaskManager options and qparams.
        qtype = options.qtype

        if qtype == "script":
            manager = flowtk.TaskManager.from_user_config()
            script = manager.qadapter.get_script_str(
                job_name="job_name", launch_dir="workdir", executable="executable",
                qout_path="qout_file.path", qerr_path="qerr_file.path",
                stdin="stdin", stdout="stdout", stderr="stderr")
            print(script)

        else:
            print(flowtk.TaskManager.autodoc())
            print("qtype supported: %s" % flowtk.all_qtypes())
            print("Use `abirun.py . manager slurm` to have the list of qparams for slurm.\n")

            if qtype is not None:
                print("QPARAMS for %s" % qtype)
                flowtk.show_qparams(qtype)

        return 0

    if options.command == "doc_scheduler":
        print("Options that can be specified in scheduler.yml:")
        print(flowtk.PyFlowScheduler.autodoc())
        return 0

    if options.command == "abibuild":
        abinit_build = flowtk.AbinitBuild()
        print()
        print(abinit_build)
        print()
        if not options.verbose:
            print("Use --verbose for additional info")
        else:
            print(abinit_build.info)
        return 0

    # Abiopen does not need a flow
    if options.command == "abiopen":
        return cli_abiopen(options, options.flowdir)

    # After this point we start to operate on the flow.
    # 0) Print logo
    # 1) Read flow from pickle file and construct nids set if needed.
    # 2) Operate on the flow depending on the options specified by the user on the CLI.
    if options.no_colors:
        # Disable colors
        termcolor.enable(False)

    if not options.no_logo:
        nrows, ncols = get_terminal_size()
        if ncols > 100: cprint(abilab.abipy_logo1(), "yellow")

        system, node, release, version, machine, processor = platform.uname()
        cprint("Running on %s -- system %s -- Python %s -- %s" % (
               gethostname(), system, platform.python_version(), "abirun" + "-" + abilab.__version__),
               'yellow', attrs=['underline'])

    wname, tname = None, None
    if options.flowdir is None:
        # Will try to figure out the location of the Flow.
        options.flowdir = os.getcwd()
    else:
        # Sometimes one wants to inspect a work or a task by just using `abirun.py flow/w0/t0 inspect`
        # without knowing its node id. flowdir_wname_tname will solve the problem!
        options.flowdir, wname, tname = flowdir_wname_tname(options.flowdir)

    # Read the flow from the pickle database.
    flow = flowtk.Flow.pickle_load(options.flowdir, remove_lock=options.remove_lock)

    # If we have selected a work/task, we have to convert wname/tname into node ids (nids)
    if wname or tname:
        if wname and tname:
            # Task
            for w_pos, work in enumerate(flow):
                if os.path.basename(work.workdir) == wname: break
            else:
                raise RuntimeError(f"Cannot find work from name: {wname}")

            for t_pos, task in enumerate(flow[w_pos]):
                if os.path.basename(task.workdir) == tname: break
            else:
                raise RuntimeError(f"Cannot find task from name: {tname}")

            # Create options.nids here
            options.nids = set([flow[w_pos].node_id, flow[w_pos][t_pos].node_id])

        else:
            # Work
            for w_pos, work in enumerate(flow):
                if os.path.basename(work.workdir) == wname: break
            else:
                raise RuntimeError(f"Cannot find work from name: {wname}")

            # Create options.nids here
            options.nids = set([flow[w_pos].node_id] + [task.node_id for task in flow[w_pos]])

    retcode = 0

    if options.command == "new_manager":
        # Read the new manager from file.
        new_manager = flowtk.TaskManager.from_file(options.manager_file)

        # Default status for new_manager is QCritical
        if options.task_status is None:
            options.task_status = Status.as_status("QCritical")

        # Change the manager of the errored tasks.
        print(f"Resetting tasks with status: {options.task_status}")
        for task in flow.iflat_tasks(status=options.task_status, nids=select_nids(flow, options)):
            task.reset()
            task.set_manager(new_manager)

        # Update the database.
        return flow.build_and_pickle_dump()

    elif options.command in ("explain_works", "explain_tasks"):
        what = options.command.split("_")[1]
        s = flow.explain(what=what, nids=select_nids(flow, options), verbose=options.verbose)
        print(s)

    elif options.command == "autoparal":
        flow.show_autoparal(nids=select_nids(flow, options), verbose=options.verbose)

    elif options.command == "panel":
        pn = abilab.abipanel()
        serve_kwargs = serve_kwargs_from_options(options)

        if options.nids is None:
            # Start Multipage app for this flow.
            from abipy.panels.flows import FlowMultiPageApp
            FlowMultiPageApp(flow, options.panel_template).serve(**serve_kwargs)

        else:
            node_list = list(flow.iflat_nodes(nids=select_nids(flow, options)))
            if len(node_list) > 1:
                print("Got more than one node in node_list:")
                print("Only the last node will be shown in the panel dashboard!\n")
                for node in node_list:
                    print(node)
                print("")

            node = node_list[-1]
            app = node.get_panel(template=options.panel_template)
            pn.serve(app, **serve_kwargs)

        return 0

    elif options.command == "events":
        flow.show_events(status=options.task_status, nids=select_nids(flow, options))

    elif options.command == "corrections":
        flow.show_corrections(status=options.task_status, nids=select_nids(flow, options))

    elif options.command == "history":
        flow.show_history(status=options.task_status, nids=select_nids(flow, options),
                          full_history=options.full_history, metadata=options.metadata)

    elif options.command == "handlers":
        if options.doc:
            flowtk.autodoc_event_handlers()
        else:
            flow.show_event_handlers(verbose=options.verbose)

    elif options.command == "single":
        cprint("abirun.py single is deprecated and will be removed in Abipy v1.0. Use `scheduler`", color="red")
        nlaunch = flow.single_shot()
        if nlaunch: flow.show_status()
        cprint("Number of tasks launched: %d" % nlaunch, "yellow")

    elif options.command == "rapid":
        cprint("abirun.py rapid is deprecated and will be removed in Abipy v1.0. Use `scheduler`", color="red")
        nlaunch = flow.rapidfire(max_nlaunch=options.max_nlaunch, max_loops=1, sleep_time=5)
        if nlaunch: flow.show_status()
        cprint("Number of tasks launched: %d" % nlaunch, "yellow")

    elif options.command == "scheduler":
        # Check that the env on the local machine is properly configured before starting the scheduler.
        abilab.abicheck()

        sched_options = {oname: getattr(options, oname) for oname in
            ("weeks", "days", "hours", "minutes", "seconds")}

        if all(v == 0 for v in sched_options.values()):
            sched = flow.make_scheduler()
        else:
            sched = flow.make_scheduler(**sched_options)

        print(sched)
        return sched.start()

    elif options.command == "status":
        # Select the method to call.
        show_func = flow.show_status if not options.summary else flow.show_summary

        if options.delay:
            flow_watch_status(flow, delay=options.delay, verbose=options.verbose,
                              nids=select_nids(flow, options), func_name=show_func.__name__)
        else:
            show_func(verbose=options.verbose, nids=select_nids(flow, options))
            if options.verbose and flow.manager.has_queue:
                print("Total number of jobs in queue: %s" % flow.manager.get_njobs_in_queue())

    elif options.command == "set_status":
        # Default status for reset is QCritical
        if options.task_status is None: options.task_status = Status.as_status("QCritical")
        new_status = Status.as_status(options.new_status)
        print("Will set all tasks with status: ", options.task_status, " to new_status", new_status)

        count = 0
        for task in flow.iflat_tasks(status=options.task_status, nids=select_nids(flow, options)):
            task.set_status(new_status, msg="Changed by abirun from %s to %s" % (task.status, new_status))
            count += 1

        print("Number of tasks modified: %s" % count)
        if count:
            # update database
            flow.pickle_dump()

    elif options.command == "open":
        flow.open_files(what=options.what, status=None, op="==", nids=select_nids(flow, options))

    elif options.command == "cancel":
        print("Number of jobs cancelled %d" % flow.cancel(nids=select_nids(flow, options)))
        # Remove directory
        if options.rmtree: flow.rmtree()

    elif options.command == "restart":
        # Default status for restart is Unconverged if no option is provided by the user.
        if options.task_status is None and options.nids is None:
            options.task_status = Status.as_status("Unconverged")

        nlaunch, excs = 0, []
        for task in flow.iflat_tasks(status=options.task_status, nids=select_nids(flow, options)):
            print("Will try to restart %s, with status %s" % (task, task.status))
            try:
                fired = task.restart()
                if fired:
                    nlaunch += 1
                    print("\tTask restarted")
            except Exception:
                excs.append(straceback())

        cprint("Total number of jobs restarted %d" % nlaunch, "blue")
        if nlaunch:
            # update database
            flow.pickle_dump()

        if excs:
            print("Exceptions raised\n")
            pprint(excs)

    elif options.command == "reset":
        if options.nids is None and options.task_status is None:
            # Default status for reset command is QCritical
            options.task_status = Status.as_status("QCritical")

        if options.task_status is not None:
            print("Resetting tasks with status: %s" % options.task_status)
        else:
            print("Resetting tasks with node ids: %s" % str(options.nids))

        count = 0
        for task in flow.iflat_tasks(status=options.task_status, nids=select_nids(flow, options)):
            print("Resetting task %s... " % task, end="")
            failed = task.reset()
            if failed:
                cprint("[FAILED]", "red")
            else:
                cprint("[OK]", "green")
                count += 1
        cprint("%d tasks have been reset" % count, "blue")

        flow.show_status()

        g = flow.find_deadlocks()
        #print("deadlocked:", gdeadlocked, "\nrunnables:", grunnables, "\nrunning:", g.running)
        if g.deadlocked and not (g.runnables or g.running):
            cprint("*** Flow is deadlocked ***", "red")

        flow.pickle_dump()

    elif options.command == "move":
        print("Moving flow to: %s ..." % options.dest)
        flow.chroot(options.dest)
        flow.move(options.dest)

    elif options.command == "tail":
        def get_path(task):
            """Helper function used to select the files of a task."""
            choices = {
                "o": task.output_file,
                "l": task.log_file,
                "e": task.stderr_file,
            }
            return getattr(choices[options.what_tail], "path")

        # Default status for tail is Running
        if options.task_status is None: options.task_status = Status.as_status("Running")

        paths = [get_path(task) for task in flow.iflat_tasks(status=options.task_status, nids=select_nids(flow, options))]

        if not paths:
            cprint("No job is running. Exiting!", "magenta")
        else:
            cprint("Press <CTRL+C> to interrupt. Number of output files %d\n" % len(paths),
                   color="magenta", end="", flush=True)
            try:
                os.system("tail -f %s" % " ".join(paths))
            except KeyboardInterrupt:
                cprint("Received KeyboardInterrupt from user\n", "yellow")

    elif options.command == "timeit":
        flow.check_status()
        time_parser = flow.parse_timing(nids=select_nids(flow, options))
        print(time_parser)
        df = time_parser.summarize()
        abilab.print_dataframe(df, title="output of time_parse.summarize():")

    #elif options.command == "squeue":
    #    for task in flow.select_tasks(nids=options.nids, wslice=options.wslice):
    #        if not task.qjob: continue
    #        print("qjob", task.qjob)
    #        print("info", task.qjob.get_info())
    #        print("estimated_start-time", task.qjob.estimated_start_time())
    #        print("qstats", task.qjob.get_stats())

    elif options.command == "deps":
        flow.check_status()
        flow.show_dependencies()

    elif options.command == "robot":
        print("Building robot for file extension:", options.robot_ext, "with task_class:", options.task_class)
        robot = abilab.Robot.from_flow(flow, outdirs="all", nids=select_nids(flow, options),
                                       ext=options.robot_ext)
        if len(robot) == 0:
            cprint("Robot couldn't find files", "yellow")
            return 1

        print("\n%s\n" % robot.get_label_files_str())

        if options.print_dataframe:
            # Print dataframe if robot provides get_dataframe method.
            if hasattr(robot, "get_dataframe"):
                try:
                    df = robot.get_dataframe()
                    abilab.print_dataframe(df, title="Output of robot.get_dataframe():", precision=4)
                except Exception as exc:
                    cprint("Exception:\n%s\n\nwhile invoking get_dataframe. Falling back to to_string" % str(exc), "red")
                    print(robot.to_string(verbose=options.verbose))
            else:
                cprint("%s does not provide `get_dataframe` method. Using `to_string`" % (
                        robot.__class__.__name__), "yellow")
                print(robot.to_string(verbose=options.verbose))

        elif options.notebook:
            # Create notebook.
            return robot.make_and_open_notebook(foreground=options.foreground)

        else:
            # Start ipython to interact with the robot.
            import IPython
            header = robot.get_label_files_str() + "\n\nType `robot` in the terminal and use <TAB> to list its methods."
            IPython.embed(header=header, robot=robot)

    elif options.command == "cycles":
        # Print cycles.
        from abipy.flowtk.abiinspect import CyclesPlotter
        cls2plotter = OrderedDict()
        for task, cycle in flow.get_task_scfcycles(nids=select_nids(flow, options),
                                                   exclude_ok_tasks=options.exclude_ok_tasks):
            print()
            cprint(repr(task), **task.status.color_opts)
            print("\n", cycle, "\n")

            label = repr(task) if options.verbose else "{} {}".format(task.__class__.__name__, task.relworkdir)
            # Could have different kind of cycles e.g. Scf, Relax, DFPT ...
            # so we group them by building multiple plotters indexed by task class.
            # Plots are produced outside of the loop.
            pkey = task.__class__.__name__
            if pkey not in cls2plotter: cls2plotter[pkey] = CyclesPlotter()
            cls2plotter[pkey].add_label_cycle(label, cycle)

        # Here comes the plot.
        if options.plot_mode is not None and cls2plotter:
            for pkey, plotter in cls2plotter.items():
                try:
                    getattr(plotter, options.plot_mode)(title="Plotter for %s class" % pkey)
                except Exception as exc:
                    cprint("Exception while invoking %s method of %s.\n%s" % (
                           options.plot_mode, plotter.__class__.__name__, str(exc)), "red")

    elif options.command == "dims":
        flow.get_dims_dataframe(nids=select_nids(flow, options),
                                printout=True, with_colors=not options.no_colors)

    elif options.command == "inspect":
        tasks = flow.select_tasks(nids=select_nids(flow, options))
        def plot_graphs():
            for task in tasks:
                if hasattr(task, "inspect"):
                    try:
                        task.inspect()
                    except Exception as exc:
                        cprint("%s: inspect method raised %s " % (task, exc), color="blue")
                else:
                    cprint("Task %s does not provide an inspect method" % task, color="blue")

        plot_graphs()

    elif options.command == "inputs":
        flow.show_inputs(varnames=options.varnames, nids=select_nids(flow, options))

    elif options.command == "abivars":
        flow.compare_abivars(varnames=options.varnames, nids=select_nids(flow, options),
                             printout=True, with_colors=not options.no_colors)

    elif options.command == "structures":
        flow.compare_structures(nids=select_nids(flow, options),
                                what=options.what,
                                verbose=options.verbose, with_spglib=False, printout=True,
                                with_colors=not options.no_colors)

    elif options.command == "ebands":
        df, ebands_plotter = flow.compare_ebands(
                                nids=select_nids(flow, options), verbose=options.verbose,
                                with_spglib=False, printout=True, with_colors=not options.no_colors)

        plot_mode = options.plot_mode
        if plot_mode is not None:
            plotfunc = getattr(ebands_plotter, plot_mode, None)
            if plotfunc is None:
                raise ValueError("Don't know how to handle plot_mode: %s" % plot_mode)
            plotfunc(tight_layout=True)

    elif options.command == "hist":
        df, robot = flow.compare_hist(nids=select_nids(flow, options), verbose=options.verbose,
                                       with_spglib=False, printout=True, with_colors=not options.no_colors)

        plot_mode = options.plot_mode
        if plot_mode is not None:
            if len(robot) == 1:
                robot.abifiles[0].plot()
            else:
                if plot_mode == "gridplot":
                    for what in robot.what_list:
                        robot.gridplot(what=what)
                elif plot_mode == "combiplot":
                    robot.combiplot()
                else:
                    raise ValueError(f"Invalid value of plot_mode: {plot_mode}")

    elif options.command == "notebook":
        return flow.write_open_notebook(options.foreground)

    elif options.command == "ipython":
        import IPython
        print("Invoking Ipython, `flow` object will be available in the Ipython terminal")
        IPython.start_ipython(argv=options.argv, user_ns={"flow": flow})

    elif options.command == "tar":
        if not options.light:
            tarfile = flow.make_tarfile(name=None,
                                        max_filesize=options.max_filesize,
                                        exclude_exts=options.exclude_exts,
                                        exclude_dirs=options.exclude_dirs,
                                        verbose=options.verbose)
            print("Created tarball file %s" % tarfile)
        else:
            tarfile = flow.make_light_tarfile()
            print("Created light tarball file %s" % tarfile)

    elif options.command == "tricky":
        flow.show_tricky_tasks(verbose=options.verbose)

    elif options.command == "debug":
        flow.debug(status=options.task_status, nids=select_nids(flow, options))

    elif options.command == "debug_reset":
        flow_debug_reset_tasks(flow, nids=select_nids(flow, options), verbose=options.verbose)

    elif options.command == "reset_jobids":
        # Make sure we have slurm.
        qtype = flow[0][0].manager.qadapter.QTYPE.lower()
        if qtype != "slurm":
            cprint("reset_jobids is only available for slurm", color="magenta", end="", flush=True)
            return 1

        # Call squeue to get list of job infos and extract slurm_jobs
        from abipy.flowtk.qutils import slurm_get_jobs
        slurm_jobs = slurm_get_jobs()

        for task in flow.iflat_tasks(status=options.task_status, nids=select_nids(flow, options)):
            qid = task.queue_id
            if qid is None: continue
            if qid not in slurm_jobs and not task.is_completed:
                print("Task:", task, "seeem to have been killed and will be automatically reset.")
                task.reset()

        return flow.build_and_pickle_dump()

    # TODO
    #elif options.command == "debug_restart":
    #    flow_debug_restart_tasks(flow, nids=select_nids(flow, options), verbose=options.verbose)

    elif options.command == "group":
        d = defaultdict(list)
        print("\nMapping `%s` ---> List of node identifiers" % options.groupby)
        for task in flow.iflat_tasks(status=options.task_status, nids=select_nids(flow, options)):
            if options.groupby == "status":
                k = task.status
            elif options.groupby == "task_class":
                k = task.__class__.__name__
            else:
                raise ValueError("Invalid groupby: `%s`" % options.groupby)
            d[k].append(task)

        for k, tasks in d.items():
            s = ", ".join([colored(str(task.node_id), **task.status.color_opts) for task in tasks])
            print("   ",k, " ---> ", s)

    elif options.command == "diff":
        if options.nids is None:
            raise ValueError("nids must be specified when using diff command")

        tasks = list(flow.iflat_tasks(nids=select_nids(flow, options)))

        if len(tasks) not in (2, 3):
            if len(tasks) == 1:
                cprint("task == task, returning\n", color="magenta", end="", flush=True)
                return 0
            else:
                raise ValueError("Don't know how to compare files produced by %d tasks" % len(tasks))

        # Build list of lists. Each sub-list contains the files associated to the i-th task.
        files_for_task = [None] * len(tasks)
        for i, task in enumerate(tasks):
            files_for_task[i] = task.select_files(options.what_diff)

        for diff_files in zip(*files_for_task):
            print("Comparing", ", ".join(os.path.relpath(p) for p in diff_files))
            args = " ".join(os.path.relpath(p) for p in diff_files)
            # TODO: I should have written a Differ object somewhere!
            os.system("vimdiff %s" % args)

    elif options.command == "networkx":
        flow.plot_networkx(mode=options.nxmode, with_edge_labels=options.edge_labels)

    elif options.command == "graphviz":
        # Select node to visualize.
        node = flow
        if wname or tname:
            if wname and tname: # Task
                node = flow[w_pos][t_pos]
            else:  # Work
                node = flow[w_pos]

        directory = tempfile.mkdtemp()
        print("Producing source files in:", directory)

        if options.dirtree:
            graph = node.get_graphviz_dirtree(engine=options.engine)
        else:
            graph = node.get_graphviz(engine=options.engine)

        # Add this liine to print the DOT string. Can be used with e.g. http://viz-js.com/
        if options.verbose: print(graph)

        graph.view(directory=directory, cleanup=False)

        if options.verbose > 1:
            # Write graph to file in png format.
            graph.format = "png"
            graph.attr(dpi=str(300))
            path = graph.render("graph", view=False, cleanup=False)
            print("Saving png file to:", path)

    elif options.command == "listext":
        if not options.listexts:
            print("\nPlease specify the file extension(s), e.g. GSR SIGRES.\nList of available extensions:\n")
            print(abilab.abiopen_ext2class_table())
            return 0

        for ext in options.listexts:
            print("")
            flow.listext(ext)
            print("")

    else:
        raise ValueError(f"Don't know what to do with command {options.command}!")

    return retcode


if __name__ == "__main__":
    sys.exit(main())



"""
This module provides objects for extracting timing data from the ABINIT output files
It also provides tools to analyze and to visualize the parallel efficiency.
"""
from __future__ import annotations

from abipy.core.mixins import NotebookWriter
from abipy.flowtk import AbinitTimerParser as _Parser

class AbinitTimerParser(_Parser, NotebookWriter):

    def yield_figs(self, **kwargs):  # pragma: no cover
        """
        This function *generates* a predefined list of matplotlib figures with minimal input from the user.
        """
        yield self.plot_stacked_hist(show=False)
        yield self.plot_efficiency(show=False)
        yield self.plot_efficiency(show=False)
        yield self.plot_pie(show=False)

    def write_notebook(self, nbpath=None) -> str:
        """
        Write an ipython notebook to nbpath. If nbpath is None, a temporary file in the current
        working directory is created. Return path to the notebook.
        """
        nbformat, nbv, nb = self.get_nbformat_nbv_nb(title=None)
        filenames = self.filenames

        nb.cells.extend([
            nbv.new_code_cell("parser = abilab.AbinitTimerParser()\nparser.parse(%s)" % str(filenames)),
            nbv.new_code_cell("display(parser.summarize())"),
            nbv.new_markdown_cell("# This is a markdown cell"),
            nbv.new_code_cell("parser.plot_stacked_hist();"),
            nbv.new_code_cell("parser.plot_efficiency(what='good', nmax=5);"),
            nbv.new_code_cell("parser.plot_efficiency(what='bad', nmax=5);"),
            nbv.new_markdown_cell("# This is a markdown cell"),
            nbv.new_code_cell("parser.plot_pie();"),

            nbv.new_code_cell("""\
for timer in parser.timers():
    print(timer)
    display(timer.get_dataframe())"""),
        ])

        return self._write_nb_nbpath(nb, nbpath)



"""
This module contains objects for analyzing
the GWAN.nc file with the e-ph vertex in the Wannier representation.

For a theoretical introduction see :cite:`Giustino2017`
"""
from __future__ import annotations

import dataclasses
import numpy as np
import pandas as pd
#import abipy.core.abinit_units as abu

from monty.string import marquee #, list_strings
from monty.functools import lazy_property
from monty.termcolor import cprint
from abipy.core.structure import Structure
from abipy.core.kpoints import kpoints_indices
from abipy.core.mixins import AbinitNcFile, Has_Structure, Has_ElectronBands, Has_Header #, NotebookWriter
from abipy.tools.typing import PathLike
from abipy.tools.numtools import BzRegularGridInterpolator, nparr_to_df
#from abipy.tools.plotting import (add_fig_kwargs, get_ax_fig_plt, get_axarray_fig_plt, set_axlims, set_visible,
#    rotate_ticklabels, ax_append_title, set_ax_xylabels, linestyles)
#from abipy.tools import duck
from abipy.electrons.ebands import ElectronBands, RobotWithEbands
#from abipy.tools.typing import Figure
from abipy.abio.robots import Robot
from abipy.eph.common import BaseEphReader


class GwanFile(AbinitNcFile, Has_Header, Has_Structure, Has_ElectronBands): # , NotebookWriter):
    """
    This file stores the e-ph matrix elements in the wannier representation
    and provides methods to analyze and plot results.

    Usage example:

    .. code-block:: python

        with GwanFile("out_GWAN.nc") as gwan:
            print(gwan)

            for spin in range(gwan.nsppol):
                # Extract the object storing the g for this spin.
                #gqk = gstore.gqk_spin[spin]
                #print(gqk)

                # Get a Dataframe with g(k, q) for all modes and bands.
                #df = gqk.get_gdf_at_qpt_kpt([1/2, 0, 0], [0, 0, 0])
                #print(df)

    .. rubric:: Inheritance Diagram
    .. inheritance-diagram:: GwanFile
    """
    @classmethod
    def from_file(cls, filepath: PathLike) -> GwanFile:
        """Initialize the object from a netcdf file."""
        return cls(filepath)

    def __init__(self, filepath: PathLike):
        super().__init__(filepath)
        self.r = GwanReader(filepath)

    @lazy_property
    def ebands(self) -> ElectronBands:
        """|ElectronBands| object."""
        return self.r.read_ebands()

    @property
    def structure(self) -> Structure:
        """|Structure| object."""
        return self.ebands.structure

    def close(self) -> None:
        """Close the file."""
        self.r.close()

    #@lazy_property
    #def gqk_spin(self) -> list:
    #    return [Gqk.from_gstore(self, spin) for spin in range(self.nsppol)]

    @lazy_property
    def params(self) -> dict:
        """dict with the convergence parameters, e.g. ``nbsum``."""
        #od = OrderedDict([
        #    ("nbsum", self.nbsum),
        #    ("nqibz", self.r.nqibz),
        #])
        ## Add EPH parameters.
        #od.update(self.r.common_eph_params)

        od = {}
        return od

    def __str__(self) -> str:
        return self.to_string()

    def to_string(self, verbose=0) -> str:
        """String representation with verbosiy level ``verbose``."""
        lines = []; app = lines.append

        app(marquee("File Info", mark="="))
        app(self.filestat(as_string=True))
        app("")
        app(self.structure.to_string(verbose=verbose, title="Structure"))

        app("")
        app(self.ebands.to_string(with_structure=False, verbose=verbose, title="Electronic Bands"))
        if verbose > 1:
            app("")
            app(self.hdr.to_string(verbose=verbose, title="Abinit Header"))

        app(f"nsppol: {self.r.nsppol}")
        app(f"gstore_completed: {bool(self.r.completed)}")
        app(f"gstore_cplex: {self.r.cplex}")
        app(f"gstore_kptopt: {self.r.kptopt}")
        app(f"gstore_qptopt: {self.r.qptopt}")
        #for spin in range(self.r.nsppol):
        #    app(f"gstore_brange_spin[{spin}]: {self.r.brange_spin[spin]}")
        #    app(f"gstore_erange_spin[{spin}]: {self.r.erange_spin[spin]}")
        #    app(f"gstore_glob_spin_nq[{spin}]: {self.r.glob_spin_nq[spin]}")

        return "\n".join(lines)

    #def write_epw_hdf5(self, filepath: PathLike) -> None:


@dataclasses.dataclass(kw_only=True)
class Gqk:
    """
    This object stores the e-ph matrix elements (g or g^2) and the matrix elements
    of the velocity operator for a given spin.
    """
    cplex: int         # 1 if |g|^2 is stored
                       # 2 if complex valued g (mind the gauge)
    spin: int          # Spin index.
    nb: int            # Number of bands
    bstart: int
    #bstop: int

    glob_nk: int       # Total number of k/q points in global matrix.
    glob_nq: int       # Note that k-points/q-points can be filtered.
                       # Use kzone, qzone and kfilter to interpret these dimensions.

    gstore: GstoreFile

    gvals: np.ndarray | None
    g2: np.ndarray | None
    vk_cart_ibz: np.ndarray | None
    vkmat_cart_ibz: np.ndarray | None

    @classmethod
    def from_gstore(cls, gstore: GstoreFile, spin: int):
        """
        Build an istance from a GstoreFile and the spin index.
        """
        ncr = gstore.r
        path = f"gqk_spin{spin+1}"
        cplex = ncr.read_dimvalue("gstore_cplex")
        nb = ncr.read_dimvalue("nb", path=path)
        glob_nk = ncr.read_dimvalue("glob_nk", path=path)
        glob_nq = ncr.read_dimvalue("glob_nq", path=path)

        # Read e-ph matrix elements
        # nctkarr_t("gvals", "dp", "gstore_cplex, nb_kq, nb_k, natom3, glob_nk, glob_nq)
        # Have to transpose the (nb_kq, nb_k) submatrix written by Fortran.
        g2, gvals = None, None
        if cplex == 1:
            g2 = ncr.read_value("gvals", path=path).transpose(0, 1, 2, 4, 3, 5).copy()

        elif cplex == 2:
            gvals = ncr.read_value("gvals", path=path).transpose(0, 1, 2, 4, 3, 5).copy()
            gvals = gvals[...,0] + 1j*gvals[...,1]

        vk_cart_ibz, vkmat_cart_ibz = None, None
        if ncr.with_vk == 1:
            # nctk_def_arrays(spin_ncid, nctkarr_t("vk_cart_ibz", "dp", "three, nb, gstore_nkibz"))
            vk_cart_ibz = ncr.read_value("vk_cart_ibz", path=path)

        if ncr.with_vk == 2:
            # Full (nb x nb) matrix.
            # Have to transpose (nb_kq, nb_k) submatrix written by Fortran.
            # nctk_def_arrays(spin_ncid, nctkarr_t("vkmat_cart_ibz", "dp", "two, three, nb, nb, gstore_nkibz"))
            vkmat_cart_ibz = ncr.read_value("vkmat_cart_ibz", path=path).transpose(0, 1, 3, 2, 4).copy()
            vkmat_cart_ibz = vkmat_cart_ibz[...,0] + 1j*vkmat_cart_ibz[...,1]

        # Note conversion between Fortran and python indexing.
        bstart = ncr.read_value("bstart", path=path) - 1
        #bstop = ncr.read_value("stop", path=path)

        data = locals()
        return cls(**{k: data[k] for k in [field.name for field in dataclasses.fields(Gqk)]})

    def __str__(self) -> str:
        return self.to_string()

    def to_string(self, verbose=0) -> str:
        """String representation with verbosiy level ``verbose``."""
        lines = []; app = lines.append

        app(marquee(f"Gqk for spin: {self.spin}", mark="="))
        app(f"cplex: {self.cplex}")
        app(f"nb: {self.nb}")
        app(f"bstart: {self.bstart}")
        app(f"glob_nk: {self.glob_nk}")
        app(f"glob_nq: {self.glob_nq}")

        return "\n".join(lines)

    @property
    def structure(self):
        return self.gstore.structure

    def get_dataframe(self, what: str = "g2") -> pd.DataFrame:
        """
        Build and return a dataframe with all the |g(k,q)|^2 if what == "g2" or
        all |v_nk|^2 if what == "v2".
        """
        if what == "g2":
            g2 = self.g2 if self.g2 is not None else np.abs(self.gvals) ** 2
            df = nparr_to_df("g2", g2, ["iq", "ik", "imode", "m_kq", "n_k"])

        elif what == "v2":
            if self.vk_cart_ibz is None:
                raise ValueError("vk_cart_ibz is not available in GSTORE!")
            # Compute the squared norm of each vector
            v2 = np.sum(self.vk_cart_ibz ** 2, axis=2)
            df = nparr_to_df("v2", v2, ["ik", "n_k"])

        else:
            raise ValueError(f"Invalid {what=}")

        #df["m_kq"] += bstart_mkq
        #df["n_k"] += bstart_nk

        return df

    def get_g2q_interpolator_kpoint(self, kpoint, method="linear", check_mesh=1):
        """
        """
        r = self.gstore.r

        # Find the index of the kpoint.
        ik_g, kpoint = r.find_ik_glob_kpoint(kpoint, self.spin)

        # Compute indices of qpoints in the ngqpt mesh.
        ngqpt, shifts = r.ngqpt, [0, 0, 0]
        q_indices = kpoints_indices(r.qbz, ngqpt, check_mesh=check_mesh)

        natom3 = 3 * len(self.structure)
        nb = self.nb
        nx, ny, nz = ngqpt

        # (glob_nq, glob_nk, natom3, m_kq, n_k)
        g2 = self.g2 if self.g2 is not None else np.abs(self.gvals) ** 2
        g2_qph_mn = g2[:,ik_g]

        # Insert g2 in g2_grid
        g2_grid = np.empty((nb, nb, natom3, nx, ny, nz))
        for nu in range(natom3):
            for g2_mn, q_inds in zip(g2_qph_mn[:,nu], q_indices):
                ix, iy, iz = q_inds
                g2_grid[:, :, nu, ix, iy, iz] = g2_mn

        return BzRegularGridInterpolator(self.structure, shifts, g2_grid, method=method)

    def get_g_qpt_kpt(self, qpoint, kpoint, what) -> np.ndarray:
        """
        Return numpy array with e-ph matrix elements the for the given (qpoint, kpoint) pair.

        Args:
            what="g2" for |g(k,q)|^2, "g" for g(k,q)
        """
        # Find the internal indices of (qpoint, kpoint)
        iq_g, qpoint = self.gstore.r.find_iq_glob_qpoint(qpoint, self.spin)
        ik_g, kpoint = self.gstore.r.find_ik_glob_kpoint(kpoint, self.spin)
        if what == "g2":
            g2 = self.g2 if self.g2 is not None else np.abs(self.gvals) ** 2
            return g2[iq_g, ik_g]
        if what == "g":
            if self.cplex != 2:
                raise ValueError("Gstore file stores g2 instead of complex g")
            return self.gvals[iq_g, ik_g]

        raise ValueError(f"Invalid {what=}")

    def get_gdf_at_qpt_kpt(self, qpoint, kpoint, what="g2") -> pd.DataFrame:
        """
        Build and return a dataframe with the |g(k,q)|^2 for the given (qpoint, kpoint) pair.

        Args:
            what="g2" for |g(k,q)|^2, "g" for g(k,q)
        """
        g2_slice = self.get_g_qpt_kpt(qpoint, kpoint, what)
        df = nparr_to_df(what, g2_slice, ["imode", "m_kq", "n_k"])
        #df["m_kq"] += bstart_mkq
        #df["n_k"] += bstart_nk

        return df

    def neq(self, other: Gqk, verbose: int) -> int:
        """
        Helper function to compare two GQK objects.
        """
        # This dimensions must agree in order to have a meaningfull comparison.
        # so raise immediately if not equal.
        aname_list = ["cplex", "spin", "nb", "glob_nk", "glob_nq"]

        for aname in aname_list:
            val1, val2 = getattr(self, aname), getattr(other, aname)

            if isinstance(val1, (str, int, float)):
                eq = val1 == val2
            elif isinstance(val1, np.ndarray):
                eq = np.allclose(val1, val2)
            else:
                raise TypeError(f"Don't know how to handle comparison for type: {type(val1)}")

            if not eq:
                raise RuntimeError(f"Different values of {aname=}, {val1=}, {val2=}")

        ierr = 0
        kws = dict(verbose=verbose) # , atol= rtol)

        # Compare v_nk or v_mn_k.
        if self.vk_cart_ibz is not None:
            if not _allclose("vk_cart_ibz", self.vk_cart_ibz, other.vk_cart_ibz, **kws): ierr += 1

        if self.vkmat_cart_ibz is not None:
            if not _allclose("vkmat_cart_ibz", self.vkmat_cart_ibz, other.vkmat_cart_ibz, **kws): ierr += 1

        # Compare g or g^2.
        if self.g2 is not None:
            if not _allclose("g2", self.g2, other.g2, **kws): ierr += 1

        if self.gvals is not None:
            if not _allclose("gvals", self.gvals, other.gvals, **kws): ierr += 1

        return ierr


class GstoreReader(BaseEphReader):
    """
    Reads data from file and constructs objects.

    .. rubric:: Inheritance Diagram
    .. inheritance-diagram:: GstoreReader
    """
    def __init__(self, filepath: PathLike):
        super().__init__(filepath)

        # Read important dimensions.
        self.nsppol = self.read_dimvalue("number_of_spins")
        self.cplex = self.read_dimvalue("gstore_cplex")
        self.nkbz = self.read_dimvalue("gstore_nkbz")
        self.nkibz = self.read_dimvalue("gstore_nkibz")
        self.nqbz = self.read_dimvalue("gstore_nqbz")
        self.nqibz = self.read_dimvalue("gstore_nqibz")

        # Read important variables.
        self.completed = self.read_value("gstore_completed")
        self.done_spin_qbz = self.read_value("gstore_done_qbz_spin")
        self.with_vk = self.read_value("gstore_with_vk")
        self.qptopt = self.read_value("gstore_qptopt")
        self.kptopt = self.read_value("kptopt")
        self.kzone = self.read_string("gstore_kzone")
        self.qzone = self.read_string("gstore_qzone")
        self.kfilter = self.read_string("gstore_kfilter")
        self.gmode = self.read_string("gstore_gmode")

        # Note conversion Fortran --> C for the isym index.
        self.brange_spin = self.read_value("gstore_brange_spin")
        self.brange_spin[:,0] -= 1
        self.erange_spin = self.read_value("gstore_erange_spin")
        # Total number of k/q points for each spin after filtering (if any)
        self.glob_spin_nq = self.read_value("gstore_glob_nq_spin")
        self.glob_nk_spin = self.read_value("gstore_glob_nk_spin")

        # K-points and q-points in the IBZ
        self.kibz = self.read_value("reduced_coordinates_of_kpoints")
        self.qibz = self.read_value("gstore_qibz")

        # K-points and q-points in the BZ
        self.kbz = self.read_value("gstore_kbz")
        self.qbz = self.read_value("gstore_qbz")
        self.ngqpt = self.read_value("gstore_ngqpt")

        # Mapping BZ --> IBZ. Note conversion Fortran --> C for the isym index.
        # nctkarr_t("gstore_kbz2ibz", "i", "six, gstore_nkbz"), &
        # nctkarr_t("gstore_qbz2ibz", "i", "six, gstore_nqbz"), &
        self.kbz2ibz = self.read_value("gstore_kbz2ibz")
        self.kbz2ibz[:,0] -= 1

        self.qbz2ibz = self.read_value("gstore_qbz2ibz")
        self.qbz2ibz[:,0] -= 1

        # Mapping q/k points in gqk --> BZ. Note conversion Fortran --> C for indexing.
        # nctkarr_t("gstore_qglob2bz", "i", "gstore_max_nq, number_of_spins"), &
        # nctkarr_t("gstore_kglob2bz", "i", "gstore_max_nk, number_of_spins") &
        self.qglob2bz = self.read_value("gstore_qglob2bz")
        self.qglob2bz -= 1
        self.kglob2bz = self.read_value("gstore_kglob2bz")
        self.kglob2bz -= 1

    def find_iq_glob_qpoint(self, qpoint, spin: int):
        """
        Find the internal index of the qpoint needed to access the gvals array.
        """
        qpoint = np.asarray(qpoint)
        for iq_g, iq_bz in enumerate(self.qglob2bz[spin]):
            if np.allclose(qpoint, self.qbz[iq_bz]):
                #print(f"Found {qpoint = } with index {iq_g = }")
                return iq_g, qpoint

        raise ValueError(f"Cannot find {qpoint = } in GSTORE.nc")

    def find_ik_glob_kpoint(self, kpoint, spin: int):
        """Find the internal indices of the kpoint needed to access the gvals array."""
        kpoint = np.asarray(kpoint)
        for ik_g, ik_bz in enumerate(self.kglob2bz[spin]):
            if np.allclose(kpoint, self.kbz[ik_bz]):
                #print(f"Found {kpoint = } with index {ik_g = }")
                return ik_g, kpoint

        raise ValueError(f"Cannot find {kpoint = } in GSTORE.nc")

    # TODO: This fix to read groups should be imported in pymatgen.
    @lazy_property
    def path2group(self) -> dict:
        return self.rootgrp.groups


class GstoreRobot(Robot, RobotWithEbands):
    """
    This robot analyzes the results contained in multiple GSTORE.nc files.

    Usage example:

    .. code-block:: python

        robot = GstoreRobot.from_files([
            "t04o_GSTORE.nc",
            "t05o_GSTORE.nc",
            ])

        robot.neq(verbose=1)

    .. rubric:: Inheritance Diagram
    .. inheritance-diagram:: GstoreRobot
    """
    EXT = "GSTORE"

    def neq(self, ref_basename: str | None = None, verbose: int = 0) -> int:
        """
        Compare all GSTORE.nc files stored in the GstoreRobot
        """
        # Find reference gstore. By default the first file in the robot is used.
        ref_gstore = self._get_ref_abifile_from_basename(ref_basename)

        exc_list = []
        ierr = 0
        for other_gstore in self.abifiles:
            if ref_gstore.filepath == other_gstore.filepath:
                continue
            print("Comparing: ", ref_gstore.basename, " with: ", other_gstore.basename)
            try:
                ierr += self._neq_two_gstores(ref_gstore, other_gstore, verbose)
                cprint("EQUAL", color="green")
            except Exception as exc:
                exc_list.append(str(exc))

        for exc in exc_list:
            cprint(exc, color="red")

        return ierr

    @staticmethod
    def _neq_two_gstores(gstore1: GstoreFile, gstore2: GstoreFile, verbose: int) -> int:
        """
        Helper function to compare two GSTORE files.
        """
        # These quantities must be the same to have a meaningfull comparison.
        aname_list = ["structure", "nsppol", "cplex", "nkbz", "nkibz",
                      "nqbz", "nqibz", "completed", "kzone", "qzone", "kfilter", "gmode",
                      "brange_spin", "erange_spin", "glob_spin_nq", "glob_nk_spin",
                     ]

        for aname in aname_list:
            self._compare_attr_name(aname, gstore1, gstore2)

        # Now compare the gkq objects for each spin.
        ierr = 0
        for spin in range(gstore1.nsppol):
            gqk1, gqk2 = gstore1.gqk_spin[spin], gstore2.gqk_spin[spin]
            ierr += gqk1.neq(gqk2, verbose)

        return ierr

    def yield_figs(self, **kwargs):  # pragma: no cover
        """
        This function *generates* a predefined list of matplotlib figures with minimal input from the user.
        Used in abiview.py to get a quick look at the results.
        """
        #for fig in self.get_ebands_plotter().yield_figs(): yield fig

    def write_notebook(self, nbpath=None) -> str:
        """
        Write a jupyter_ notebook to ``nbpath``. If nbpath is None, a temporary file in the current
        working directory is created. Return path to the notebook.
        """
        nbformat, nbv, nb = self.get_nbformat_nbv_nb(title=None)

        args = [(l, f.filepath) for l, f in self.items()]
        nb.cells.extend([
            #nbv.new_markdown_cell("# This is a markdown cell"),
            nbv.new_code_cell("robot = abilab.GstoreRobot(*%s)\nrobot.trim_paths()\nrobot" % str(args)),
            #nbv.new_code_cell("ebands_plotter = robot.get_ebands_plotter()"),
        ])

        # Mixins
        #nb.cells.extend(self.get_baserobot_code_cells())
        #nb.cells.extend(self.get_ebands_code_cells())

        return self._write_nb_nbpath(nb, nbpath)



"""Database with the names of the input variables used in Abinit and in other main programs."""
from __future__ import annotations

import sys

from collections import OrderedDict

##############
# Public API #
##############


def get_abinit_variables():
    """Returns the database with the description of the ABINIT variables."""
    from abipy.abio.abivar_database.variables import get_codevars
    return get_codevars()["abinit"]


def get_anaddb_variables():
    """Returns the database with the description of the ANADDB variables."""
    from abipy.abio.abivar_database.variables import get_codevars
    return get_codevars()["anaddb"]


def docvar(varname, executable="abinit"):
    """Return the `Variable` object associated to this name."""
    from abipy.abio.abivar_database.variables import get_codevars
    return get_codevars()[executable][varname]


def abinit_help(varname: str, info=True, stream=sys.stdout) -> None:
    """
    Print the abinit documentation on the ABINIT input variable `varname`
    """
    database = get_abinit_variables()
    if hasattr(varname, "abivarname"): varname = varname.name
    try:
        var = database[varname]
    except KeyError:
        return stream.write("Variable %s not in database" % varname)

    text = "## Default value: %s\n## Description:\n\n%s\n" % (str(var.defaultval), var.text)

    if info: text += str(var.info)
    text = text.replace("[[", "\033[1m").replace("]]", "\033[0m")

    # FIXME: There are unicode chars in abinit doc (Greek symbols)
    try:
        stream.write(text)
    except UnicodeEncodeError:
        stream.write(text.encode('ascii', 'ignore'))
    stream.write("\n")


def repr_html_from_abinit_string(text: str) -> str:
    """
    Given a string `text` with an Abinit input file, replace all variables
    with HTML links pointing to the official documentation. 
    Return new string.
    """
    var_database = get_abinit_variables()

    # https://stackoverflow.com/questions/6116978/python-replace-multiple-strings
    # define desired replacements here e.g. rep = {"condition1": "", "condition2": "text"}
    # ordered dict and sort by length is needed because variable names can overlap e.g. kpt, kptopt pair
    import re
    rep = {vname: var.html_link(label=vname) for vname, var in var_database.items()}
    rep = OrderedDict([(re.escape(k), rep[k]) for k in sorted(rep.keys(), key=lambda n: len(n), reverse=True)])

    # Use these three lines to do the replacement
    pattern = re.compile("|".join(rep.keys()))
    text = pattern.sub(lambda m: rep[re.escape(m.group(0))], text)
    return text.replace("\n", "<br>")



# coding: utf-8
"""
Objects common to the other eph modules.
"""
from __future__ import annotations

import numpy as np
import abipy.core.abinit_units as abu

from collections import OrderedDict
from monty.functools import lazy_property
from abipy.electrons.ebands import ElectronsReader

# Phonon frequency in Ha below which e-ph matrix elements are set to zero.
EPH_WTOL = 1e-6


class BaseEphReader(ElectronsReader):
    """
    Provides methods common to the netcdf files produced by the EPH code.
    See Abinit docs for the meaning of the variables.
    """

    @lazy_property
    def ddb_ngqpt(self):
        """Q-Mesh for DDB file."""
        return self.read_value("ddb_ngqpt")

    @lazy_property
    def ngqpt(self):
        """Effective Q-mesh used in to compute integrals (ph_linewidts, e-ph self-energy)."""
        return self.read_value("ngqpt")

    @lazy_property
    def ph_ngqpt(self):
        """Q-mesh for Phonon DOS, interpolated A2F ..."""
        return self.read_value("ph_ngqpt")

    @lazy_property
    def eph_ngqpt_fine(self):
        """Q-mesh for interpolated DFPT potentials"""
        return self.read_value("eph_ngqpt_fine")

    @lazy_property
    def common_eph_params(self):
        """
        Read basic parameters (scalars) from the netcdf files produced by the EPH code and cache them
        """
        od = OrderedDict([
            ("ddb_nqbz", np.prod(self.ddb_ngqpt)),
            ("eph_nqbz_fine", np.prod(self.eph_ngqpt_fine)),
            ("ph_nqbz", np.prod(self.ph_ngqpt)),
        ])

        for vname in ["eph_intmeth", "eph_fsewin", "eph_fsmear", "eph_extrael", "eph_fermie"]:
            value = self.read_value(vname)
            if vname in ("eph_intmeth",):
                value = int(value)
            else:
                value = float(value)
            od[vname] = value

        return od


def glr_frohlich(qpoint, becs_cart, epsinf_cart, phdispl_cart_bohr, phfreqs_ha, structure,
                 qdamp=None, eph_wtol=EPH_WTOL, tol_qnorm=1e-6):
    """
    Compute the long-range part of the e-ph matrix element with the simplified Frohlich model
    i.e. we include only G = 0 and the <k+q,b1|e^{i(q+G).r}|b2,k> coefficient is replaced by delta_{b1, b2}

    Args:
        qpoint: |Kpoint| object.
        becs_cart: (natom, 3, 3) arrays with Born effective charges in Cartesian coordinates.
        epsinf_cart: (3, 3) array with macroscopic dielectric tensor in Cartesian coordinates.
        phdispl_cart_bohr: (natom3_nu, natom3) complex array with phonon displacement in Cartesian coordinates (Bohr)
        phfreqs_ha: (3 * natom) array with phonon frequencies in Ha.
        structure: |Structure| object.
        qdamp: Exponential damping.
        eph_wtol: Set g to zero below this phonon frequency.
        tol_qnorm: Tolerance of the norm of the q-point.

    Return:
        (natom3) complex array with glr_nu.
    """
    natom = len(structure)
    natom3 = natom * 3
    # 2pi factor already included in frac_coords
    qcs = qpoint.cart_coords
    q_eps_q = np.dot(qcs, np.matmul(epsinf_cart, qcs))
    phdispl_cart_bohr = np.reshape(phdispl_cart_bohr, (natom3, natom, 3))

    xred = structure.frac_coords
    # Acoustic modes are included --> assume BECS fullfill charge neutrality
    glr_nu = np.empty(natom3, dtype=complex)
    for nu in range(natom3):
        if phfreqs_ha[nu] < EPH_WTOL or q_eps_q < tol_qnorm: continue
        num = 0.0j
        for iat in range(natom):
            cdd = phdispl_cart_bohr[nu, iat] * np.exp(-2.0j * np.pi * np.dot(qpoint.frac_coords, xred[iat]))
            num += np.dot(qcs, np.matmul(becs_cart[iat], cdd))
        glr_nu[nu] = num / (q_eps_q * np.sqrt(2.0 * phfreqs_ha[nu]))

    fact = 1
    if qdamp is not None: fact = np.exp(-qpoint.norm**2/(4*qdamp))
    return fact * glr_nu * 4j * np.pi / (structure.volume*abu.Ang_Bohr**3)



"""
This module contains objects for postprocessing polaron calculations
using the results stored in the VARPEQ.nc file.

For a theoretical introduction see ...
"""
from __future__ import annotations

import dataclasses
import numpy as np
import pandas as pd
import abipy.core.abinit_units as abu

from collections import defaultdict
from monty.string import marquee #, list_strings
from monty.functools import lazy_property
from monty.termcolor import cprint
from abipy.core.func1d import Function1D
from abipy.core.structure import Structure
from abipy.core.kpoints import kpoints_indices, kmesh_from_mpdivs, map_grid2ibz, IrredZone
from abipy.core.mixins import AbinitNcFile, Has_Structure, Has_ElectronBands, NotebookWriter
from abipy.tools.typing import PathLike
from abipy.tools.plotting import (add_fig_kwargs, get_ax_fig_plt, get_axarray_fig_plt, set_axlims, set_visible,
    rotate_ticklabels, ax_append_title, set_ax_xylabels, linestyles, Marker, set_grid_legend)
from abipy.electrons.ebands import ElectronBands, RobotWithEbands
from abipy.dfpt.phonons import PhononBands
from abipy.dfpt.ddb import DdbFile
from abipy.tools.typing import Figure
from abipy.tools.numtools import BzRegularGridInterpolator, gaussian
from abipy.iotools import bxsf_write
from abipy.abio.robots import Robot
from abipy.eph.common import BaseEphReader


#TODO Finalize the implementation. Look at Pedro's implementation for GFR
#from abipy.electrons.effmass_analyzer import EffMassAnalyzer
#
#class FrohlichAnalyzer:
#
#    def __init__(self, gsr_kpath, ddb, verbose = 0, **anaddb_kwargs):
#        """
#        """
#        ebands_kpath = ElectronBands.as_ebands(gsr_kpath)
#        self.ddb = DdbFile.as_ddb(ddb)
#        self.verbose = verbose
#
#        r = self.ddb.anaget_epsinf_and_becs(verbose=verbose, return_input=True, **anaddb_kwargs)
#        self.epsinf, self.becs = r.epsinf, r.becs
#
#        self.diel_gen, diel_inp = self.ddb.anaget_dielectric_tensor_generator(verbose=verbose, return_input=True, **anaddb_kwargs)
#        self.eps0_tensor = self.diel_gen.tensor_at_frequency(0.0)
#
#        # Spherical average of eps_inf and eps_0 (real part only)
#        einf_savg, e0_savg = self.epsinf.trace() / 3, self.eps0_tensor.real.trace() / 3
#
#        self.kappa = 1 / (1/einf_savg - 1/e0_savg)
#
#        self.emana = EffMassAnalyzer(ebands_kpath)
#
#    def __str__(self) -> str:
#        return self.to_string()
#
#    def to_string(self, verbose: int = 0) -> str:
#        """String representation with verbosity level verbose"""
#        lines = []
#        app = lines.append
#
#        app("epsilon_infinity in Cartesian coordinates:")
#        app(str(self.epsinf))
#        app("BECS:")
#        app(str(self.becs))
#        app("eps0 tensor:")
#        app(str(self.eps0_tensor))
#        app(f"kappa = {self.kappa}")
#
#        return "\n".join(lines)
#
#    def analyze_band_edges(self):
#        self.emana.select_cbm()
#        self.emana.summarize()
#        self.emana.plot_emass()
#        self.emana.select_vbm()
#        self.emana.summarize()
#        self.emana.plot_emass()
#        #self.emana.select_band_edges()


@dataclasses.dataclass(kw_only=True)
class Entry:
    name: str
    latex: str
    info: str
    utype: str
    #color: str


# NB: All quantities are in atomic units!
_ALL_ENTRIES = [
    Entry(name="E_pol", latex=r'$E_{pol}$', info="Formation energy", utype="energy"),
    Entry(name="E_el", latex=r'$E_{el}$', info="Electronic part", utype="energy"),
    Entry(name="E_ph", latex=r'$E_{ph}$', info="Phonon part", utype="energy"),
    Entry(name="elph", latex=r'$E_{elph}$', info="e-ph term", utype="energy"),
    Entry(name="epsilon", latex=r"$\varepsilon$", info="Polaron eigenvalue", utype="energy"),
    Entry(name="gsr", latex=r"$|\nabla|$", info="||gradient||", utype="gradient"),
]

# Convert to dictionary: name --> Entry
_ALL_ENTRIES = {e.name: e for e in _ALL_ENTRIES}


class VarpeqFile(AbinitNcFile, Has_Structure, Has_ElectronBands, NotebookWriter):
    """
    This file stores the results of a VARPEQ calculations: SCF cycle, A_nk, B_qnu
    and provides methods to analyze and plot results.

    Usage example:

    .. code-block:: python

        from abipy.eph.varpeq import VarpeqFile
        with VarpeqFile("out_VARPEQ.nc") as varpeq:
            print(varpeq)
            varpeq.plot_scf_cycle()

    .. rubric:: Inheritance Diagram
    .. inheritance-diagram:: VarpeqFile
    """

    @classmethod
    def from_file(cls, filepath: PathLike) -> VarpeqFile:
        """Initialize the object from a netcdf file."""
        return cls(filepath)

    def __init__(self, filepath: PathLike):
        super().__init__(filepath)
        self.r = VarpeqReader(filepath)

    @lazy_property
    def ebands(self) -> ElectronBands:
        """|ElectronBands| object."""
        return self.r.read_ebands()

    @property
    def structure(self) -> Structure:
        """|Structure| object."""
        return self.ebands.structure

    def close(self) -> None:
        """Close the file."""
        self.r.close()

    @lazy_property
    def polaron_spin(self) -> list[Polaron]:
        """List of polaron objects, one for each spin (if any)."""
        return [Polaron.from_varpeq(self, spin) for spin in range(self.r.nsppol)]

    @lazy_property
    def params(self) -> dict:
        """dict with the convergence parameters, e.g. ``nbsum``."""
        r = self.r

        ksampling = self.ebands.kpoints.ksampling
        ngkpt, shifts = ksampling.mpdivs, ksampling.shifts
        nkbz = np.prod(ngkpt)

        od = dict([
            ("nkbz", nkbz),
            ("ngkpt", ngkpt),
            ("invsc_size", 1.0 / (nkbz * ((abu.Ang_Bohr * self.structure.lattice.volume) ** (1/3)))),
            ("frohl_ntheta", r.frohl_ntheta),
        ])
        return od

    def __str__(self) -> str:
        return self.to_string()

    def to_string(self, verbose=0) -> str:
        """String representation with verbosiy level ``verbose``."""
        lines = []; app = lines.append

        app(marquee("File Info", mark="="))
        app(self.filestat(as_string=True))
        app("")
        app(self.structure.to_string(verbose=verbose, title="Structure"))
        app("")
        app(self.ebands.to_string(with_structure=False, verbose=verbose, title="Electronic Bands"))

        app("")
        app("VARPEQ parameters:")
        app(f"varpeq_pkind: {self.r.varpeq_pkind}")
        #app(f"gstore_cplex: {self.r.cplex}")
        #app(f"gstore_kzone: {self.r.kzone}")
        #app(f"gstore_kfilter: {self.r.kfilter}")
        #app(f"gstore_kptopt: {self.r.kptopt}")
        #app(f"gstore_qptopt: {self.r.qptopt}")

        for spin in range(self.nsppol):
            polaron = self.polaron_spin[spin]
            df = polaron.get_final_results_df()
            app(f"Last SCF iteration. Energies in eV units")
            app(str(df))
            app("")

        return "\n".join(lines)

    def yield_figs(self, **kwargs):  # pragma: no cover
        """
        This function *generates* a predefined list of matplotlib figures with minimal input from the user.
        """
        for polaron in self.polaron_spin:
            yield polaron.plot_scf_cycle(show=False)

    def write_notebook(self, nbpath=None) -> str:
        """
        Write a jupyter_ notebook to ``nbpath``. If nbpath is None, a temporay file in the current
        working directory is created. Return path to the notebook.
        """
        nbformat, nbv, nb = self.get_nbformat_nbv_nb(title=None)

        nb.cells.extend([
            nbv.new_code_cell("varpeq = abilab.abiopen('%s')" % self.filepath),
            nbv.new_code_cell("print(varpeq)"),
        ])

        return self._write_nb_nbpath(nb, nbpath)


@dataclasses.dataclass(kw_only=True)
class Polaron:
    """
    This object stores the polaron coefficients A_kn, B_qnu for a given spin and all the nstate polaron states.
    Provides methods to plot |A_nk|^2 or |B_qnu|^2 together with the band structures (fatbands-like plots).
    """
    spin: int          # Spin index.
    nstates: int       # Number of polaronic states.
    nb: int            # Number of bands in A_kn,
    nk: int            # Number of k-points in A_kn, (including filtering if any)
    nq: int            # Number of q-points in B_qnu (including filtering if any)
    bstart: int        # First band starts at bstart
    bstop: int         # Last band (python convention)
    varpeq: VarpeqFile

    @classmethod
    def from_varpeq(cls, varpeq: VarpeqFile, spin: int) -> Polaron:
        """
        Build an istance from a VarpeqFile and the spin index.
        """
        r = varpeq.r
        nstates, nk, nq, nb = r.nstates, r.nk_spin[spin], r.nq_spin[spin], r.nb_spin[spin]
        bstart, bstop = r.brange_spin[spin]

        data = locals()
        return cls(**{k: data[k] for k in [field.name for field in dataclasses.fields(Polaron)]})

    @property
    def structure(self) -> Structure:
        """Crystalline structure."""
        return self.varpeq.structure

    @property
    def ebands(self) -> ElectronBands:
        """Electron bands."""
        return self.varpeq.ebands

    @lazy_property
    def kpoints(self) -> np.ndarray:
        """Reduced coordinates of the k-points."""
        return self.varpeq.r.read_value("kpts_spin")[self.spin, :self.nk]

    @lazy_property
    def qpoints(self) -> np.ndarray:
        """Reduced coordinates of the q-points."""
        return self.varpeq.r.read_value("qpts_spin")[self.spin, :self.nq]

    @lazy_property
    def a_kn(self) -> np.ndarray:
        """A_{pnk} coefficients for this spin."""
        return self.varpeq.r.read_value("a_spin", cmode="c")[self.spin, :self.nstates, :self.nk, :self.nb]

    @lazy_property
    def b_qnu(self) -> np.ndarray:
        """B_{pqnu} coefficients for this spin."""
        return self.varpeq.r.read_value("b_spin", cmode="c")[self.spin, :self.nstates, :self.nq]

    @lazy_property
    def scf_df_state(self) -> list[pd.DataFrame]:
        """
        List of dataframes with the SCF iterations. One dataframe for each polaron.
        NB: Energies are converted from Ha to eV.
        """
        # int nstep2cv_spin(nsppol, nstates) ;
        #   Number of steps to convergence at each state for each spin
        # double scf_hist_spin(nsppol, nstates, nstep, six) ;
        #   SCF optimization history at each state for each spin
        # int cvflag_spin(nsppol, nstates) ;
        #   0 --> calculation is not converged
        #   1 --> calculation is converged
        spin = self.spin
        r = self.varpeq.r
        nstep2cv = r.read_variable("nstep2cv_spin")[spin]
        scf_hist = r.read_variable("scf_hist_spin")[spin]
        cvflag = r.read_variable("cvflag_spin")[spin]

        def ufact_k(k):
            """Convert energies to eV"""
            entry = _ALL_ENTRIES[k]
            if entry.utype == "energy":
                return abu.Ha_eV
            if entry.utype == "gradient":
                return 1.0
            raise ValueError(f"Don't know how to convert {entry=}")

        # Build list of dataframe.
        df_list = []
        for pstate in range(self.nstates):
            n = nstep2cv[pstate]
            dct = {k: scf_hist[pstate, :n, i] * ufact_k(k) for i, k in enumerate(_ALL_ENTRIES)}
            df = pd.DataFrame(dct)
            # Add metadata to the attrs dictionary
            df.attrs["converged"] = bool(cvflag[pstate])
            df_list.append(df)

        return df_list

    def get_final_results_df(self, with_params: bool=False) -> pd.DataFrame:
        """
        Return daframe with the last iteration for all polaronic states.
        NB: Energies are in eV.
        """
        row_list = []
        for pstate in range(self.nstates):
            df = self.scf_df_state[pstate]
            row = {"pstate": pstate, "spin": self.spin}
            row.update(df.iloc[-1].to_dict())
            row["converged"] = df.attrs["converged"]
            if with_params:
                row.update(self.varpeq.params)
            row_list.append(row)

        return pd.DataFrame(row_list)

    def __str__(self) -> str:
        return self.to_string()

    def to_string(self, verbose: int=0) -> str:
        """
        String representation with verbosiy level verbose.
        """
        lines = []; app = lines.append

        app(marquee(f"Ank for spin: {self.spin}", mark="="))
        app(f"nstates: {self.nstates}")
        app(f"nk: {self.nk}")
        app(f"nq: {self.nq}")
        app(f"nb: {self.nb}")
        app(f"bstart: {self.bstart}, bstop: {self.bstop}")
        ksampling = self.ebands.kpoints.ksampling
        ngkpt, shifts = ksampling.mpdivs, ksampling.shifts
        app(f"ksampling: {str(ksampling)}")
        ngqpt = self.varpeq.r.ngqpt
        app(f"q-mesh: {ngqpt}")
        if verbose:
            for pstate in range(self.nstates):
                app("For %d: 1/N_k sum_nk |A_nk|^2: %f" % (pstate, np.sum(np.abs(self.a_kn[pstate]) ** 2) / self.nk))

        return "\n".join(lines)

    @lazy_property
    def ngkpt_and_shifts(self) -> tuple:
        """
        Return k-mesh divisions and shifts.
        """
        ksampling = self.ebands.kpoints.ksampling
        ngkpt, shifts = ksampling.mpdivs, ksampling.shifts

        if ngkpt is None:
            raise ValueError("Non diagonal k-meshes are not supported!")
        if len(shifts) > 1:
            raise ValueError("Multiple k-shifts are not supported!")

        return ngkpt, shifts

    def get_title(self, with_gaps: bool=True) -> str:
        """
        Return string with title for matplotlib plots.
        """
        varpeq = self.varpeq
        pre = "" if varpeq.ebands.nsppol == 1 else f"spin={self.spin}"
        if not with_gaps:
            return f"{pre}{varpeq.r.varpeq_pkind} polaron"
        else:
            gaps_string = varpeq.ebands.get_gaps_string()
            return f"{pre}{varpeq.r.varpeq_pkind} polaron, {gaps_string}"

    def insert_a_inbox(self, fill_value=None) -> tuple:
        """
        Return a_data, ngkpt, shifts where a_data is a
        (nstates, nb, nkx, nky, nkz)) array with A_{pnk} with p the polaron index.
        """
        # Need to know the shape of the k-mesh.
        ngkpt, shifts = self.ngkpt_and_shifts
        k_indices = kpoints_indices(self.kpoints, ngkpt)
        nx, ny, nz = ngkpt

        shape = (self.nstates, self.nb, nx, ny, nz)
        a_data = np.empty(shape, dtype=complex) if fill_value is None else np.full(shape, fill_value, dtype=complex)

        for ip in range(self.nstates):
            for ib in range(self.nb):
                for a_cplx, k_inds in zip(self.a_kn[ip,:,ib], k_indices, strict=True):
                    ix, iy, iz = k_inds
                    a_data[ip, ib, ix, iy, iz] = a_cplx

        return a_data, ngkpt, shifts

    def insert_b_inbox(self, fill_value=None) -> tuple:
        """
        Return b_data, ngqpt, shifts where b_data is a
        (nstates, nb, nqx, nqy, nqz)) array with B_{pqnu} with p the polaron index.
        """
        # Need to know the shape of the q-mesh (always Gamma-centered)
        ngqpt, shifts = self.varpeq.r.ngqpt, [0, 0, 0]
        q_indices = kpoints_indices(self.qpoints, ngqpt)

        natom3 = 3 * len(self.structure)
        nx, ny, nz = ngqpt

        shape = (self.nstates, natom3, nx, ny, nz)
        b_data = np.empty(shape, dtype=complex) if fill_value is None else np.full(shape, fill_value, dtype=complex)

        for ip in range(self.nstates):
            for nu in range(natom3):
                for b_cplx, q_inds in zip(self.b_qnu[ip,:,nu], q_indices, strict=True):
                    ix, iy, iz = q_inds
                    b_data[ip, nu, ix, iy, iz] = b_cplx

        return b_data, ngqpt, shifts

    def get_a2_interpolator_state(self) -> BzRegularGridInterpolator:
        """
        Build and return an interpolator for |A_nk|^2 for each polaronic state.
        """
        a_data, ngkpt, shifts = self.insert_a_inbox()

        return [BzRegularGridInterpolator(self.structure, shifts, np.abs(a_data[pstate]) ** 2, method="linear")
                for pstate in range(self.nstates)]

    def get_b2_interpolator_state(self) -> BzRegularGridInterpolator:
        """
        Build and return an interpolator for |B_qnu|^2 for each polaronic state.
        """
        b_data, ngqpt, shifts = self.insert_b_inbox()

        return [BzRegularGridInterpolator(self.structure, shifts, np.abs(b_data[pstate]) ** 2, method="linear")
                for pstate in range(self.nstates)]

    def write_a2_bxsf(self, filepath: PathLike. fill_value=0.0) -> None:
        r"""
        Export \sum_n |A_{pnk}|^2 in BXSF format suitable for visualization with xcrysden (use ``xcrysden --bxsf FILE``).
        Requires gamma-centered k-mesh.

        Args:
            filepath: BXSF filename.
        """
        # NB: the kmesh must be gamma-centered, multiple shifts are not supported.
        # Init values with 0. This is relevant only if kfiltering is being used.
        a_data, ngkpt, shifts = self.insert_a_inbox(fill_value=fill_value)

        # Compute \sum_n A^2_{pnk}.
        a2_data = (a_data.abs() ** 2).sum(index=1)
        fermie = a2_data.mean()

        bxsf_write(filepath, self.structure, 1, self.nstates, ngkpt, a2_data, fermie, unit="Ha")

    def write_b2_bxsf(self, filepath: PathLike, fill_value=0.0) -> None:
        r"""
        Export \sum_{\nu} |B_{q\nu}|^2 in BXSF format suitable for visualization with xcrysden (use ``xcrysden --bxsf FILE``).

        Args:
            filepath: BXSF filename.
        """
        # NB: qmesh must be gamma-centered, multiple shifts are not supported.
        # Init values with 0. This is relevant only if kfiltering is being used.
        b_data, ngqpt, shifts = self.insert_b_inbox(fill_value=fill_value)

        # Compute \sum_{\nu} B^2_{pq\nu}.
        b2_data = (b_data.abs() ** 2).sum(index=1)
        fermie = b2_data.mean()

        bxsf_write(filepath, self.structure, 1, self.nstates, ngqpt, b2_data, fermie, unit="Ha")

    #@add_fig_kwargs
    #def plot_bz_sampling(self, what="kpoints", fold=False,
    #                     ax=None, pmg_path=True, with_labels=True, **kwargs) -> Figure:
    #    """
    #    Plots a 3D representation of the Brillouin zone with the sampling.

    #    Args:
    #        what: "kpoints" or "qpoints"
    #        fold: whether the points should be folded inside the first Brillouin Zone.
    #            Defaults to False.
    #    """
    #    bz_points = dict(kpoints=self.kpoints, qpoints=self.qpoints)[what]
    #    kws = dict(ax=ax, pmg_path=pmg_path, with_labels=with_labels, fold=fold, kpoints=bz_points)

    #    return self.structure.plot_bz(show=False, **kws)

    @add_fig_kwargs
    def plot_scf_cycle(self, ax_mat=None, fontsize=8, **kwargs) -> Figure:
        """
        Plot the SCF cycle.

        Args:
            ax_max: |matplotlib-Axes| or None if a new figure should be created.
            fontsize: fontsize for legends and titles
        """
        # Build grid of plots.
        nrows, ncols = self.nstates, 2
        ax_mat, fig, plt = get_axarray_fig_plt(ax_mat, nrows=nrows, ncols=ncols,
                                               sharex=False, sharey=False, squeeze=False)

        for pstate in range(self.nstates):
            df = self.scf_df_state[pstate]
            niter = len(df)
            xs = np.arange(1, niter + 1)

            for iax, ax in enumerate(ax_mat[pstate]):
                # Create a twin Axes sharing the x-axis
                #grad_ax = ax
                grad_ax = ax.twinx()

                for ilab, (name, entry) in enumerate(_ALL_ENTRIES.items()):
                    # Convert energies to Hartree. Keep gradient as it is.
                    ys = df[name].to_numpy()

                    _ax, energy_like = ax, True
                    if entry.utype == "gradient":
                        _ax, energy_like = grad_ax, False

                    if iax == 0:
                        # Plot values linear scale.
                        _ax.plot(xs, ys, label=entry.latex)
                    else:
                        # Plot deltas in logscale.
                        _ax.plot(xs, np.abs(ys - ys[-1]), label=entry.latex)
                        _ax.set_yscale("log")

                _ax.set_xlim(1, niter)

                if energy_like:
                    ylabel = "Energy (eV)" if iax == 0 else r"$|\Delta E|$ (eV)"
                else:
                    ylabel = r"$|\nabla|$" if iax == 0 else r"$|\Delta |\nabla|$"

                set_grid_legend(_ax, fontsize, xlabel="Iteration") #, ylabel=ylabel)
                _ax.set_ylabel(ylabel)

        fig.suptitle(self.get_title(with_gaps=True))
        fig.tight_layout()

        return fig

    @add_fig_kwargs
    def plot_ank_with_ebands(self, ebands_kpath,
                             ebands_kmesh=None, lpratio: int=5, method="gaussian", step: float=0.05, width: float=0.1,
                             nksmall: int=20, normalize: bool=False, with_title=True,
                             ax_mat=None, ylims=None, scale=10, marker_color="gold", fontsize=12, **kwargs) -> Figure:
        """
        Plot electron bands with markers with size proportional to |A_nk|^2.

        Args:
            ebands_kpath: ElectronBands or Abipy file providing an electronic band structure along a k-path.
            ebands_kmesh: ElectronBands or Abipy file providing an electronic band structure with k in the IBZ.
            nksmall:
            normalize: Rescale the two DOS to plot them on the same scale.
            lpratio: Ratio between the number of star functions and the number of ab-initio k-points.
                The default should be OK in many systems, larger values may be required for accurate derivatives.
            method: Integration scheme for DOS
            step: Energy step (eV) of the linear mesh for DOS computation.
            width: Standard deviation (eV) of the gaussian for DOS computation.
            with_title: True to add title with chemical formula and gaps.
            ax_mat: List of |matplotlib-Axes| or None if a new figure should be created.
            ylims: Set the data limits for the y-axis. Accept tuple e.g. ``(left, right)``
                   or scalar e.g. ``left``. If left (right) is None, default values are used.
            scale: Scaling factor for |A_nk|^2.
            marker_color: Color for markers.
            fontsize: fontsize for legends and titles
        """
        nrows, ncols = self.nstates, 2
        gridspec_kw = {'width_ratios': [2, 1]}
        ax_mat, fig, plt = get_axarray_fig_plt(ax_mat, nrows=nrows, ncols=ncols,
                                               sharex=False, sharey=True, squeeze=False, gridspec_kw=gridspec_kw)
        # Get interpolators for A_nk
        a2_interp_state = self.get_a2_interpolator_state()

        # DEBUG SECTION
        #ref_kn = np.abs(self.a_kn) ** 2
        #for ik, kpoint in enumerate(self.kpoints):
        #    interp = a2_interp_state[0].eval_kpoint(kpoint)
        #    print("MAX (A2 ref - A2 interp) at qpoint", kpoint)
        #    print((np.abs(ref_kn[ik] - interp)).max())

        df = self.get_final_results_df()

        ebands_kpath = ElectronBands.as_ebands(ebands_kpath)
        ymin, ymax = +np.inf, -np.inf
        for pstate in range(self.nstates):
            x, y, s = [], [], []
            for ik, kpoint in enumerate(ebands_kpath.kpoints):
                enes_n = ebands_kpath.eigens[self.spin, ik, self.bstart:self.bstop]
                for e, a2 in zip(enes_n, a2_interp_state[pstate].eval_kpoint(kpoint), strict=True):
                    x.append(ik); y.append(e); s.append(scale * a2)
                    ymin, ymax = min(ymin, e), max(ymax, e)

            points = Marker(x, y, s, color=marker_color, edgecolors='gray', alpha=0.8, label=r'$|A_{n\mathbf{k}}|^2$')
            ax = ax_mat[pstate, 0]
            ebands_kpath.plot(ax=ax, points=points, show=False)
            ax.legend(loc="best", shadow=True, fontsize=fontsize)

            # Add energy and convergence status
            with_info = True
            if with_info:
                data = (df[df["pstate"] == pstate]).to_dict(orient="list")
                e_pol_ev, converged = float(data["E_pol"][0]), bool(data["converged"][0])
                title = f"Formation energy: {e_pol_ev:.3f} eV, {converged=}"
                ax.set_title(title, fontsize=8)

            if pstate != self.nstates - 1:
                set_visible(ax, False, *["legend", "xlabel"])

        vertices_names = [(k.frac_coords, k.name) for k in ebands_kpath.kpoints]

        if ebands_kmesh is None:
            print(f"Computing ebands_kmesh with star-function interpolation and {nksmall=} ...")
            edos_ngkpt = self.structure.calc_ngkpt(nksmall)
            r = self.ebands.interpolate(lpratio=lpratio, vertices_names=vertices_names, kmesh=edos_ngkpt)
            ebands_kmesh = r.ebands_kmesh

        # Get electronic DOS from ebands_kmesh.
        edos_kws = dict(method=method, step=step, width=width)
        edos = ebands_kmesh.get_edos(**edos_kws)
        edos_mesh = edos.spin_dos[self.spin].mesh
        e0 = self.ebands.fermie

        ##################
        # Compute A_nk DOS
        ##################
        # NB: A_nk does not necessarily have the symmetry of the lattice so we have to loop over the full BZ.
        # Here we get the mapping BZ --> IBZ needed to obtain the KS eigenvalues e_nk from the IBZ for the DOS.
        kmesh = ebands_kmesh.get_bz2ibz_bz_points()

        for pstate in range(self.nstates):
            # Compute A^2(E) DOS with A_nk in the full BZ
            ank_dos = np.zeros(len(edos_mesh))
            for ik_ibz, kpoint in zip(kmesh.bz2ibz, kmesh.bz_kpoints):
                enes_n = ebands_kmesh.eigens[self.spin, ik_ibz, self.bstart:self.bstop]
                a2_n = a2_interp_state[pstate].eval_kpoint(kpoint)
                for e, a2 in zip(enes_n, a2_n):
                    ank_dos += a2 * gaussian(edos_mesh, width, center=e-e0)
            ank_dos /= np.product(kmesh.ngkpt)
            ank_dos = Function1D(edos_mesh, ank_dos)
            print(f"For {pstate=}, A^2(E) integrates to:", ank_dos.integral_value, " Ideally, it should be 1.")

            ax = ax_mat[pstate, 1]
            edos_opts = {"color": "black",} if self.spin == 0 else {"color": "red"}
            edos.plot_ax(ax, e0, spin=self.spin, normalize=normalize, exchange_xy=True, label="eDOS(E)", **edos_opts)
            ank_dos.plot_ax(ax, exchange_xy=True, normalize=normalize, label=r"$A^2$(E)", color=marker_color)

            # Computes A2(E) using only k-points in the IBZ. This is just for testing.
            # A2_IBZ(E) should be equal to A2(E) only if A_nk fullfills the lattice symmetries. See notes above.
            with_ibz_a2dos = True
            if with_ibz_a2dos:
                ank_dos = np.zeros(len(edos_mesh))
                for ik_ibz, kpoint in enumerate(ebands_kmesh.kpoints):
                    weight = kpoint.weight
                    enes_n = ebands_kmesh.eigens[self.spin, ik_ibz, self.bstart:self.bstop]
                    for e, a2 in zip(enes_n, a2_interp_state[pstate].eval_kpoint(kpoint), strict=True):
                        ank_dos += weight * a2 * gaussian(edos_mesh, width, center=e-e0)
                ank_dos = Function1D(edos_mesh, ank_dos)
                print(f"For {pstate=}, A2_IBZ(E) integrates to:", ank_dos.integral_value, " Ideally, it should be 1.")
                ank_dos.plot_ax(ax, exchange_xy=True, normalize=normalize, label=r"$A^2_{IBZ}$(E)", color=marker_color, ls="--")

            set_grid_legend(ax, fontsize, xlabel="Arb. unit")

            if pstate != self.nstates - 1:
                set_visible(ax, False, *["legend", "xlabel"])

        if ylims is None:
            # Automatic ylims.
            ymin -= 0.1 * abs(ymin)
            ymax += 0.1 * abs(ymax)
            ylims = [ymin - e0, ymax - e0]

        for ax in ax_mat.ravel():
            set_axlims(ax, ylims, "y")

        if with_title:
            fig.suptitle(self.get_title(with_gaps=True))

        return fig

    @add_fig_kwargs
    def plot_bqnu_with_ddb(self, ddb, with_phdos=True, anaddb_kwargs=None, **kwargs) -> Figure:
        """
        High-level interface to plot phonon energies with markers with size proportional to |B_qnu|^2.
        Similar to plot_bqnu_with_phbands but this function receives in input a DdbFile or a
        path to a ddb file and automates the computation of the phonon bands by invoking anaddb.

        Args:
            ddb: DdbFile or path to file.
            with_phdos: True if phonon DOS should be computed and plotter.
            anaddb_kwargs: Optional arguments passed to anaddb.
            kwargs: Optional arguments passed to plot_bqnu_with_phbands.
        """
        ddb = DdbFile.as_ddb(ddb)
        anaddb_kwargs = {} if anaddb_kwargs is None else anaddb_kwargs

        with ddb.anaget_phbst_and_phdos_files(**anaddb_kwargs) as g:
            phbst_file, phdos_file = g[0], g[1]
            phbands_qpath = phbst_file.phbands
            return self.plot_bqnu_with_phbands(phbands_qpath,
                                               phdos_file=phdos_file if with_phdos else None,
                                               ddb=ddb, **kwargs)

    @add_fig_kwargs
    def plot_bqnu_with_phbands(self, phbands_qpath, phdos_file=None,
                               ddb=None, width=0.001, normalize: bool=True,
                               verbose=0, anaddb_kwargs=None, with_title=True,
                               ax_mat=None, scale=10, marker_color="gold", fontsize=12, **kwargs) -> Figure:
        """
        Plot phonon energies with markers with size proportional to |B_qnu|^2.

        Args:
            phbands_qpath: PhononBands or Abipy file providing a phonon band structure.
            phdos_file:
            ddb: DdbFile or path to file.
            width: Standard deviation (eV) of the gaussian.
            normalize: Rescale the two DOS to plot them on the same scale.
            verbose:
            anaddb_kwargs: Optional arguments passed to anaddb.
            with_title: True to add title with chemical formula and gaps.
            ax_mat: List of |matplotlib-Axes| or None if a new figure should be created.
            scale: Scaling factor for |B_qnu|^2.
            marker_color: Color for markers.
            fontsize: fontsize for legends and titles.
        """
        with_phdos = phdos_file is not None  and ddb is not None
        nrows, ncols, gridspec_kw = self.nstates, 1, None
        if with_phdos:
            ncols, gridspec_kw = 2, {'width_ratios': [2, 1]}

        ax_mat, fig, plt = get_axarray_fig_plt(ax_mat, nrows=nrows, ncols=ncols,
                                               sharex=False, sharey=True, squeeze=False, gridspec_kw=gridspec_kw)

        phbands_qpath = PhononBands.as_phbands(phbands_qpath)

        # Get interpolators for B_qnu
        b2_interp_state = self.get_b2_interpolator_state()

        for pstate in range(self.nstates):
            x, y, s = [], [], []
            for iq, qpoint in enumerate(phbands_qpath.qpoints):
                omegas_nu = phbands_qpath.phfreqs[iq,:]
                for w, b2 in zip(omegas_nu, b2_interp_state[pstate].eval_kpoint(qpoint), strict=True):
                    x.append(iq); y.append(w); s.append(scale * b2)

            ax = ax_mat[pstate, 0]
            points = Marker(x, y, s, color=marker_color, edgecolors='gray', alpha=0.8, label=r'$|B_{\nu\mathbf{q}}|^2$')
            phbands_qpath.plot(ax=ax, points=points, show=False)
            ax.legend(loc="best", shadow=True, fontsize=fontsize)

            if pstate != self.nstates - 1:
                set_visible(ax, False, *["legend", "xlabel"])

        if not with_phdos:
            if with_title: fig.suptitle(self.get_title(with_gaps=True))
            return fig

        ####################
        # Compute B_qnu DOS
        ####################
        # NB: The B_qnu do not necessarily have the symmetry of the lattice so we have to loop over the full BZ.
        # Add phdos and |B_qn| dos. Mesh is given in eV, values are in states/eV.
        phdos = phdos_file.phdos
        phdos_ngqpt = np.diagonal(phdos_file.qptrlatt) # Use same q-mesh as phdos
        phdos_shifts = [0.0, 0.0, 0.0]
        phdos_nqbz = np.product(phdos_ngqpt)
        wmesh = phdos.mesh

        # Here we get the mapping BZ --> IBZ needed to obtain the ph frequencies omega_qnu from the IBZ for the DOS.
        bz_qpoints = kmesh_from_mpdivs(phdos_ngqpt, phdos_shifts)
        #bz2ibz = map_grid2ibz(self.structure, ibz_qpoints, phdos_ngqpt, has_timrev=True)

        #ibz = IrredZone.from_ngkpt(self.structure, phdos_ngqpt, phdos_shifts, kptopt=1)

        # Call anaddb (again) to get phonons on the nqpt mesh.
        anaddb_kwargs = {} if anaddb_kwargs is None else anaddb_kwargs
        phbands_qmesh = ddb.anaget_phmodes_at_qpoints(qpoints=bz_qpoints, ifcflag=1, verbose=verbose, **anaddb_kwargs)
        if len(phbands_qmesh.qpoints) != np.product(phdos_ngqpt):
            raise RuntimeError(f"{len(phbands_qmesh.qpoints)=} != {np.product(phdos_ngqpt)=}")
        #print(phbands_qmesh.qpoints)

        for pstate in range(self.nstates):
            # TODO New version using the BZ. Requires new VARPEQ.nc file with all symmetries
            # The B_qnu do not necessarily have the symmetry of the lattice so we have to loop over the full BZ.
            # Get mapping BZ --> IBZ needed to obtain the KS eigenvalues e_nk from the IBZ for the DOS
            """
            bqnu_dos = np.zeros(len(wmesh))
            for iq_ibz, qpoint in zip(bz2ibz, bz_qpoints):
                freqs_nu = phbands_qmesh.phfreqs[iq_ibz]
                for w, b2 in zip(freqs_nu, b2_interp_state[pstate].eval_kpoint(qpoint), strict=True)
                    bqnu_dos += b2 gaussian(wmesh, width, center=w)
            bqnu_dos /= np.product(phdos_ngqpt)
            """

            # Compute B2(E) using only q-points in the IBZ. This is just for testing.
            # B2_IBZ(E) should be equal to B2(E) only if B_qnu fullfill the lattice symmetries. See notes above.
            #with_ibz_b2dos = True
            #if with_ibz_b2dos:
            bqnu_dos = np.zeros(len(wmesh))
            for iq, qpoint in enumerate(phbands_qmesh.qpoints):
                weight = qpoint.weight
                #print(weight, 1.0/phdos_nqbz)
                freqs_nu = phbands_qmesh.phfreqs[iq]
                for w, b2 in zip(freqs_nu, b2_interp_state[pstate].eval_kpoint(qpoint), strict=True):
                    bqnu_dos += weight * b2 * gaussian(wmesh, width, center=w)

            bqnu_dos = Function1D(wmesh, bqnu_dos)

            ax = ax_mat[pstate, 1]
            phdos.plot_ax(ax, exchange_xy=True, normalize=normalize, label="phDOS(E)", color="black")
            bqnu_dos.plot_ax(ax, exchange_xy=True, normalize=normalize, label=r"$B^2$(E)", color=marker_color)
            set_grid_legend(ax, fontsize, xlabel="Arb. unit")

            if pstate != self.nstates - 1:
                set_visible(ax, False, *["legend", "xlabel"])

        if with_title:
            fig.suptitle(self.get_title(with_gaps=True))

        return fig


class VarpeqReader(BaseEphReader):
    """
    Reads data from file and constructs objects.

    .. rubric:: Inheritance Diagram
    .. inheritance-diagram:: VarpeqReader
    """

    def __init__(self, filepath: PathLike):
        super().__init__(filepath)

        # Netcdf Variables

        # int eph_task ;
        # int nkbz ;
        # int nqbz ;
        # int frohl_ntheta ;
        # double varpeq_tolgrs ;
        # double e_frohl ;
        # char varpeq_pkind(fnlen) ;
        # char varpeq_aseed(fnlen) ;
        # int ngkpt(three) ;
        # int gstore_ngqpt(three) ;
        # int nk_spin(nsppol) ;
        # int nq_spin(nsppol) ;
        # int nb_spin(nsppol) ;
        # int brange_spin(nsppol, two) ;
        # int cvflag_spin(nsppol, nstates) ;
        # int nstep2cv_spin(nsppol, nstates) ;
        # double scf_hist_spin(nsppol, nstates, nstep, six) ;
        # double kpts_spin(nsppol, max_nk, three) ;
        # double qpts_spin(nsppol, max_nq, three) ;
        # double cb_min_spin(nsppol) ;
        # double vb_max_spin(nsppol) ;
        # double a_spin(nsppol, nstates, max_nk, max_nb, two) ;
        # double b_spin(nsppol, nstates, max_nq, natom3, two) ;

        # Read important dimensions.
        self.nsppol = self.read_dimvalue("nsppol")
        self.nstates = self.read_dimvalue("nstates")
        self.nk_spin = self.read_value("nk_spin")
        self.nq_spin = self.read_value("nq_spin")
        #self.nkbz = self.read_dimvalue("nkbz")
        #self.nqbz = self.read_dimvalue("nqbz")
        self.varpeq_pkind = self.read_string("varpeq_pkind")
        #self.varpeq_aseed = self.read_string("varpeq_aseed")
        self.ngqpt = self.read_value("gstore_ngqpt")
        self.frohl_ntheta = self.read_value("frohl_ntheta")

        # Read important variables.
        #self.completed = self.read_value("gstore_completed")
        #self.done_spin_qbz = self.read_value("gstore_done_qbz_spin")
        #self.qptopt = self.read_value("gstore_qptopt")
        #self.kptopt = self.read_value("kptopt")
        #self.kzone = self.read_string("gstore_kzone")
        #self.qzone = self.read_string("gstore_qzone")
        #self.kfilter = self.read_string("gstore_kfilter")
        #self.gmode = self.read_string("gstore_gmode")

        # Note conversion Fortran --> C for the isym index.
        self.brange_spin = self.read_value("brange_spin")
        self.brange_spin[:,0] -= 1
        self.nb_spin = self.brange_spin[:,1] - self.brange_spin[:,0]

        #self.erange_spin = self.read_value("gstore_erange_spin")
        # Total number of k/q points for each spin after filtering (if any)
        #self.glob_spin_nq = self.read_value("gstore_glob_nq_spin")
        #self.glob_nk_spin = self.read_value("gstore_glob_nk_spin")


class VarpeqRobot(Robot, RobotWithEbands):
    """
    This robot analyzes the results contained in multiple VARPEQ.nc files.

    Usage example:

    .. code-block:: python

        robot = VarpeqRobot.from_files([
            "out1_VARPEQ.nc",
            "out2_VARPEQ.nc",
            ])

        print(robot)
        df = robot.get_final_results_df()

    .. rubric:: Inheritance Diagram
    .. inheritance-diagram:: VarpeqRobot
    """

    EXT = "VARPEQ"

    def __str__(self) -> str:
        return self.to_string()

    def to_string(self, verbose=0) -> str:
        """String representation with verbosiy level ``verbose``."""
        lines = []; app = lines.append
        df = self.get_final_results_df()
        lines.append(str(df))

        return "\n".join(lines)

    def get_final_results_df(self, spin=None, sortby=None, with_params: bool=True) -> pd.DataFrame:
        """
        Return dataframe with the last iteration for all polaronic states.
        NB: Energies are in eV.

        Args:
            spin:
            sortby: Name to sort by.
            with_params:
        """
        df_list = []
        for abifile in self.abifiles:
            if spin is None:
                for polaron in abifile.polaron_spin:
                    df_list.append(polaron.get_final_results_df(with_params=with_params))
            else:
                polaron = abifile.polaron_spin[spin]
                df_list.append(polaron.get_final_results_df(with_params=with_params))

        df = pd.concat(df_list)
        if sortby and sortby in df: df = df.sort_values(sortby)
        return df

    @add_fig_kwargs
    def plot_kconv(self, colormap="jet", fontsize=12, **kwargs) -> Figure:
        """
        Plot the convergence of the results wrt to the k-point sampling.

        Args:
            colormap: Color map. Have a look at the colormaps here and decide which one you like:
            fontsize: fontsize for legends and titles
        """
        nsppol = self.getattr_alleq("nsppol")

        # Build grid of plots.
        nrows, ncols = len(_ALL_ENTRIES), nsppol
        ax_mat, fig, plt = get_axarray_fig_plt(None, nrows=nrows, ncols=ncols,
                                               sharex=True, sharey=False, squeeze=False)
        cmap = plt.get_cmap(colormap)
        for spin in range(nsppol):
            df = self.get_final_results_df(spin=spin, sortby=None)
            xs = df["invsc_size"]
            xvals = np.linspace(0.0, 1.1 * xs.max(), 100)

            for ix, ylabel in enumerate(_ALL_ENTRIES):
                ax = ax_mat[ix, spin]
                ys = df[ylabel]

                # Plot ab-initio points.
                ax.scatter(xs, ys, color="red", marker="o")

                # Plot fit using the first nn points.
                for nn in range(1, len(xs)):
                    color = cmap((nn - 1) / len(xs))
                    p = np.poly1d(np.polyfit(xs[:nn+1], ys[:nn+1], deg=1))
                    ax.plot(xvals, p(xvals), color=color, ls="--")

                xlabel = "Inverse supercell size (Bohr$^-1$)" if ix == len(_ALL_ENTRIES) - 1 else None
                set_grid_legend(ax, fontsize, xlabel=xlabel, ylabel=f"{ylabel} (eV)", legend=False)
                ax.tick_params(axis='x', color='black', labelsize='20', pad=5, length=5, width=2)

        return fig

    def yield_figs(self, **kwargs):  # pragma: no cover
        """
        This function *generates* a predefined list of matplotlib figures with minimal input from the user.
        Used in abiview.py to get a quick look at the results.
        """
        #yield self.plot_scf_cycle(show=False)
        yield self.plot_kconv()

    def write_notebook(self, nbpath=None) -> str:
        """
        Write a jupyter_ notebook to ``nbpath``. If nbpath is None, a temporary file in the current
        working directory is created. Return path to the notebook.
        """
        nbformat, nbv, nb = self.get_nbformat_nbv_nb(title=None)

        args = [(l, f.filepath) for l, f in self.items()]
        nb.cells.extend([
            nbv.new_code_cell("robot = abilab.VarpeqRobot(*%s)\nrobot.trim_paths()\nrobot" % str(args)),
            #nbv.new_code_cell("ebands_plotter = robot.get_ebands_plotter()"),
        ])

        return self._write_nb_nbpath(nb, nbpath)
